from numpy import dtype

from ..utils import docval, getargs
from ..spec.spec import DtypeHelper

__all__ = [
    "Error",
    "DtypeError",
    "MissingError",
    "ExpectedArrayError",
    "ShapeError",
    "MissingDataType",
    "IllegalLinkError",
    "IncorrectDataType"
]


class Error:

    @docval({'name': 'name', 'type': str, 'doc': 'the name of the component that is erroneous'},
            {'name': 'reason', 'type': str, 'doc': 'the reason for the error'},
            {'name': 'location', 'type': str, 'doc': 'the location of the error', 'default': None})
    def __init__(self, **kwargs):
        self.__name = getargs('name', kwargs)
        self.__reason = getargs('reason', kwargs)
        self.__location = getargs('location', kwargs)
        if self.__location is not None:
            self.__str = "%s (%s): %s" % (self.__name, self.__location, self.__reason)
        else:
            self.__str = "%s: %s" % (self.name, self.reason)

    @property
    def name(self):
        return self.__name

    @property
    def reason(self):
        return self.__reason

    @property
    def location(self):
        return self.__location

    @location.setter
    def location(self, loc):
        self.__location = loc
        self.__str = "%s (%s): %s" % (self.__name, self.__location, self.__reason)

    def __str__(self):
        return self.__str

    def __repr__(self):
        return self.__str__()


class DtypeError(Error):

    @docval({'name': 'name', 'type': str, 'doc': 'the name of the component that is erroneous'},
            {'name': 'expected', 'type': (dtype, type, str, list), 'doc': 'the expected dtype'},
            {'name': 'received', 'type': (dtype, type, str, list), 'doc': 'the received dtype'},
            {'name': 'location', 'type': str, 'doc': 'the location of the error', 'default': None})
    def __init__(self, **kwargs):
        name = getargs('name', kwargs)
        expected = getargs('expected', kwargs)
        received = getargs('received', kwargs)
        if isinstance(expected, list):
            expected = DtypeHelper.simplify_cpd_type(expected)
        reason = "incorrect type - expected '%s', got '%s'" % (expected, received)
        loc = getargs('location', kwargs)
        super().__init__(name, reason, location=loc)


class MissingError(Error):
    @docval({'name': 'name', 'type': str, 'doc': 'the name of the component that is erroneous'},
            {'name': 'location', 'type': str, 'doc': 'the location of the error', 'default': None})
    def __init__(self, **kwargs):
        name = getargs('name', kwargs)
        reason = "argument missing"
        loc = getargs('location', kwargs)
        super().__init__(name, reason, location=loc)


class MissingDataType(Error):
    @docval({'name': 'name', 'type': str, 'doc': 'the name of the component that is erroneous'},
            {'name': 'data_type', 'type': str, 'doc': 'the missing data type'},
            {'name': 'location', 'type': str, 'doc': 'the location of the error', 'default': None})
    def __init__(self, **kwargs):
        name, data_type = getargs('name', 'data_type', kwargs)
        self.__data_type = data_type
        reason = "missing data type %s" % self.__data_type
        loc = getargs('location', kwargs)
        super().__init__(name, reason, location=loc)

    @property
    def data_type(self):
        return self.__data_type


class ExpectedArrayError(Error):

    @docval({'name': 'name', 'type': str, 'doc': 'the name of the component that is erroneous'},
            {'name': 'expected', 'type': (tuple, list), 'doc': 'the expected shape'},
            {'name': 'received', 'type': str, 'doc': 'the received data'},
            {'name': 'location', 'type': str, 'doc': 'the location of the error', 'default': None})
    def __init__(self, **kwargs):
        name = getargs('name', kwargs)
        expected = getargs('expected', kwargs)
        received = getargs('received', kwargs)
        reason = "incorrect shape - expected an array of shape '%s', got non-array data '%s'" % (expected, received)
        loc = getargs('location', kwargs)
        super().__init__(name, reason, location=loc)


class ShapeError(Error):

    @docval({'name': 'name', 'type': str, 'doc': 'the name of the component that is erroneous'},
            {'name': 'expected', 'type': (tuple, list), 'doc': 'the expected shape'},
            {'name': 'received', 'type': (tuple, list), 'doc': 'the received shape'},
            {'name': 'location', 'type': str, 'doc': 'the location of the error', 'default': None})
    def __init__(self, **kwargs):
        name = getargs('name', kwargs)
        expected = getargs('expected', kwargs)
        received = getargs('received', kwargs)
        reason = "incorrect shape - expected '%s', got '%s'" % (expected, received)
        loc = getargs('location', kwargs)
        super().__init__(name, reason, location=loc)


class IllegalLinkError(Error):
    """
    A validation error for indicating that a link was used where an actual object
    (i.e. a dataset or a group) must be used
    """

    @docval({'name': 'name', 'type': str, 'doc': 'the name of the component that is erroneous'},
            {'name': 'location', 'type': str, 'doc': 'the location of the error', 'default': None})
    def __init__(self, **kwargs):
        name = getargs('name', kwargs)
        reason = "illegal use of link"
        loc = getargs('location', kwargs)
        super().__init__(name, reason, location=loc)


class IncorrectDataType(Error):
    """
    A validation error for indicating that the incorrect data_type (not dtype) was used.
    """

    @docval({'name': 'name', 'type': str, 'doc': 'the name of the component that is erroneous'},
            {'name': 'expected', 'type': str, 'doc': 'the expected data_type'},
            {'name': 'received', 'type': str, 'doc': 'the received data_type'},
            {'name': 'location', 'type': str, 'doc': 'the location of the error', 'default': None})
    def __init__(self, **kwargs):
        name = getargs('name', kwargs)
        expected = getargs('expected', kwargs)
        received = getargs('received', kwargs)
        reason = "incorrect data_type - expected '%s', got '%s'" % (expected, received)
        loc = getargs('location', kwargs)
        super().__init__(name, reason, location=loc)



from hdmf import Data, Container
from hdmf.common import get_type_map
from hdmf.testing import TestCase


class TestCommonTypeMap(TestCase):

    def test_base_types(self):
        tm = get_type_map()
        cls = tm.get_container_cls('hdmf-common', 'Container')
        self.assertIs(cls, Container)
        cls = tm.get_container_cls('hdmf-common', 'Data')
        self.assertIs(cls, Data)



from .spec import NAME_WILDCARD
from .spec import Spec
from .spec import AttributeSpec
from .spec import DtypeSpec
from .spec import DtypeHelper
from .spec import RefSpec
from .spec import DatasetSpec
from .spec import LinkSpec
from .spec import GroupSpec
from .catalog import SpecCatalog
from .namespace import SpecNamespace
from .namespace import NamespaceCatalog
from .namespace import SpecReader
from .write import NamespaceBuilder
from .write import SpecWriter
from .write import export_spec

from ..utils import docval



import numpy as np
from abc import abstractmethod, ABCMeta


class Array:

    def __init__(self, data):
        self.__data = data
        if hasattr(data, 'dtype'):
            self.dtype = data.dtype
        else:
            tmp = data
            while isinstance(tmp, (list, tuple)):
                tmp = tmp[0]
            self.dtype = type(tmp)

    @property
    def data(self):
        return self.__data

    def __len__(self):
        return len(self.__data)

    def get_data(self):
        return self.__data

    def __getidx__(self, arg):
        return self.__data[arg]

    def __sliceiter(self, arg):
        return (x for x in range(*arg.indices(len(self))))

    def __getitem__(self, arg):
        if isinstance(arg, list):
            idx = list()
            for i in arg:
                if isinstance(i, slice):
                    idx.extend(x for x in self.__sliceiter(i))
                else:
                    idx.append(i)
            return np.fromiter((self.__getidx__(x) for x in idx), dtype=self.dtype)
        elif isinstance(arg, slice):
            return np.fromiter((self.__getidx__(x) for x in self.__sliceiter(arg)), dtype=self.dtype)
        elif isinstance(arg, tuple):
            return (self.__getidx__(arg[0]), self.__getidx__(arg[1]))
        else:
            return self.__getidx__(arg)


class AbstractSortedArray(Array, metaclass=ABCMeta):
    '''
    An abstract class for representing sorted array
    '''

    @abstractmethod
    def find_point(self, val):
        pass

    def get_data(self):
        return self

    def __lower(self, other):
        ins = self.find_point(other)
        return ins

    def __upper(self, other):
        ins = self.__lower(other)
        while self[ins] == other:
            ins += 1
        return ins

    def __lt__(self, other):
        ins = self.__lower(other)
        return slice(0, ins)

    def __le__(self, other):
        ins = self.__upper(other)
        return slice(0, ins)

    def __gt__(self, other):
        ins = self.__upper(other)
        return slice(ins, len(self))

    def __ge__(self, other):
        ins = self.__lower(other)
        return slice(ins, len(self))

    @staticmethod
    def __sort(a):
        if isinstance(a, tuple):
            return a[0]
        else:
            return a

    def __eq__(self, other):
        if isinstance(other, list):
            ret = list()
            for i in other:
                eq = self == i
                ret.append(eq)
            ret = sorted(ret, key=self.__sort)
            tmp = list()
            for i in range(1, len(ret)):
                a, b = ret[i-1], ret[i]
                if isinstance(a, tuple):
                    if isinstance(b, tuple):
                        if a[1] >= b[0]:
                            b[0] = a[0]
                        else:
                            tmp.append(slice(*a))
                    else:
                        if b > a[1]:
                            tmp.append(slice(*a))
                        elif b == a[1]:
                            a[1] == b+1
                        else:
                            ret[i] = a
                else:
                    if isinstance(b, tuple):
                        if a < b[0]:
                            tmp.append(a)
                    else:
                        if b - a == 1:
                            ret[i] = (a, b)
                        else:
                            tmp.append(a)
            if isinstance(ret[-1], tuple):
                tmp.append(slice(*ret[-1]))
            else:
                tmp.append(ret[-1])
            ret = tmp
            return ret
        elif isinstance(other, tuple):
            ge = self >= other[0]
            ge = ge.start
            lt = self < other[1]
            lt = lt.stop
            if ge == lt:
                return ge
            else:
                return slice(ge, lt)
        else:
            lower = self.__lower(other)
            upper = self.__upper(other)
            d = upper - lower
            if d == 1:
                return lower
            elif d == 0:
                return None
            else:
                return slice(lower, upper)

    def __ne__(self, other):
        eq = self == other
        if isinstance(eq, tuple):
            return [slice(0, eq[0]), slice(eq[1], len(self))]
        else:
            return [slice(0, eq), slice(eq+1, len(self))]


class SortedArray(AbstractSortedArray):
    '''
    A class for wrapping sorted arrays. This class overrides
    <,>,<=,>=,==, and != to leverage the sorted content for
    efficiency.
    '''

    def __init__(self, array):
        super().__init__(array)

    def find_point(self, val):
        return np.searchsorted(self.data, val)


class LinSpace(SortedArray):

    def __init__(self, start, stop, step):
        self.start = start
        self.stop = stop
        self.step = step
        self.dtype = float if any(isinstance(s, float) for s in (start, stop, step)) else int
        self.__len = int((stop - start)/step)

    def __len__(self):
        return self.__len

    def find_point(self, val):
        nsteps = (val-self.start)/self.step
        fl = int(nsteps)
        if fl == nsteps:
            return int(fl)
        else:
            return int(fl+1)

    def __getidx__(self, arg):
        return self.start + self.step*arg



import os
from h5py import File
import numpy as np
from abc import ABCMeta, abstractmethod

from hdmf.query import HDMFDataset, Query
from hdmf.array import SortedArray, LinSpace
from hdmf.testing import TestCase


class AbstractQueryMixin(metaclass=ABCMeta):

    @abstractmethod
    def getDataset(self):
        raise NotImplementedError('Cannot run test unless getDataset is implemented')

    def setUp(self):
        self.dset = self.getDataset()
        self.wrapper = HDMFDataset(self.dset)

    def test_get_dataset(self):
        array = self.wrapper.dataset
        self.assertIsInstance(array, SortedArray)

    def test___gt__(self):
        '''
        Test wrapper greater than magic method
        '''
        q = self.wrapper > 5
        self.assertIsInstance(q, Query)
        result = q.evaluate()
        expected = [False, False, False, False, False,
                    False, True, True, True, True]
        expected = slice(6, 10)
        self.assertEqual(result, expected)

    def test___ge__(self):
        '''
        Test wrapper greater than or equal magic method
        '''
        q = self.wrapper >= 5
        self.assertIsInstance(q, Query)
        result = q.evaluate()
        expected = [False, False, False, False, False,
                    True, True, True, True, True]
        expected = slice(5, 10)
        self.assertEqual(result, expected)

    def test___lt__(self):
        '''
        Test wrapper less than magic method
        '''
        q = self.wrapper < 5
        self.assertIsInstance(q, Query)
        result = q.evaluate()
        expected = [True, True, True, True, True,
                    False, False, False, False, False]
        expected = slice(0, 5)
        self.assertEqual(result, expected)

    def test___le__(self):
        '''
        Test wrapper less than or equal magic method
        '''
        q = self.wrapper <= 5
        self.assertIsInstance(q, Query)
        result = q.evaluate()
        expected = [True, True, True, True, True,
                    True, False, False, False, False]
        expected = slice(0, 6)
        self.assertEqual(result, expected)

    def test___eq__(self):
        '''
        Test wrapper equals magic method
        '''
        q = self.wrapper == 5
        self.assertIsInstance(q, Query)
        result = q.evaluate()
        expected = [False, False, False, False, False,
                    True, False, False, False, False]
        expected = 5
        self.assertTrue(np.array_equal(result, expected))

    def test___ne__(self):
        '''
        Test wrapper not equal magic method
        '''
        q = self.wrapper != 5
        self.assertIsInstance(q, Query)
        result = q.evaluate()
        expected = [True, True, True, True, True,
                    False, True, True, True, True]
        expected = [slice(0, 5), slice(6, 10)]
        self.assertTrue(np.array_equal(result, expected))

    def test___getitem__(self):
        '''
        Test wrapper getitem using slice
        '''
        result = self.wrapper[0:5]
        expected = [0, 1, 2, 3, 4]
        self.assertTrue(np.array_equal(result, expected))

    def test___getitem__query(self):
        '''
        Test wrapper getitem using query
        '''
        q = self.wrapper < 5
        result = self.wrapper[q]
        expected = [0, 1, 2, 3, 4]
        self.assertTrue(np.array_equal(result, expected))


class SortedQueryTest(AbstractQueryMixin, TestCase):

    path = 'SortedQueryTest.h5'

    def getDataset(self):
        self.f = File(self.path, 'w')
        self.input = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        self.d = self.f.create_dataset('dset', data=self.input)
        return SortedArray(self.d)

    def tearDown(self):
        self.f.close()
        if os.path.exists(self.path):
            os.remove(self.path)


class LinspaceQueryTest(AbstractQueryMixin, TestCase):

    path = 'LinspaceQueryTest.h5'

    def getDataset(self):
        return LinSpace(0, 10, 1)


class CompoundQueryTest(TestCase):

    def getM(self):
        return SortedArray(np.arange(10, 20, 1))

    def getN(self):
        return SortedArray(np.arange(10.0, 20.0, 0.5))

    def setUp(self):
        self.m = HDMFDataset(self.getM())
        self.n = HDMFDataset(self.getN())

    # TODO: test not completed
    # def test_map(self):
    #     q = self.m == (12, 16)                # IN operation
    #     q.evaluate()                          # [2,3,4,5]
    #     q.evaluate(False)                     # RangeResult(2,6)
    #     r = self.m[q]                         # noqa: F841
    #     r = self.m[q.evaluate()]              # noqa: F841
    #     r = self.m[q.evaluate(False)]         # noqa: F841

    def tearDown(self):
        pass



import json

from hdmf.spec import GroupSpec, DatasetSpec, AttributeSpec
from hdmf.testing import TestCase


class GroupSpecTests(TestCase):
    def setUp(self):
        self.attributes = [
            AttributeSpec('attribute1', 'my first attribute', 'text'),
            AttributeSpec('attribute2', 'my second attribute', 'text')
        ]

        self.dset1_attributes = [
            AttributeSpec('attribute3', 'my third attribute', 'text'),
            AttributeSpec('attribute4', 'my fourth attribute', 'text')
        ]

        self.dset2_attributes = [
            AttributeSpec('attribute5', 'my fifth attribute', 'text'),
            AttributeSpec('attribute6', 'my sixth attribute', 'text')
        ]

        self.datasets = [
            DatasetSpec('my first dataset',
                        'int',
                        name='dataset1',
                        attributes=self.dset1_attributes,
                        linkable=True),
            DatasetSpec('my second dataset',
                        'int',
                        name='dataset2',
                        attributes=self.dset2_attributes,
                        linkable=True,
                        data_type_def='VoltageArray')
        ]

        self.subgroups = [
            GroupSpec('A test subgroup',
                      name='subgroup1',
                      linkable=False),
            GroupSpec('A test subgroup',
                      name='subgroup2',
                      linkable=False)

        ]
        self.ndt_attr_spec = AttributeSpec('data_type', 'the data type of this object', 'text', value='EphysData')
        self.ns_attr_spec = AttributeSpec('namespace', 'the namespace for the data type of this object',
                                          'text', required=False)

    def test_constructor(self):
        spec = GroupSpec('A test group',
                         name='root_constructor',
                         groups=self.subgroups,
                         datasets=self.datasets,
                         attributes=self.attributes,
                         linkable=False)
        self.assertFalse(spec['linkable'])
        self.assertListEqual(spec['attributes'], self.attributes)
        self.assertListEqual(spec['datasets'], self.datasets)
        self.assertNotIn('data_type_def', spec)
        self.assertIs(spec, self.subgroups[0].parent)
        self.assertIs(spec, self.subgroups[1].parent)
        self.assertIs(spec, self.attributes[0].parent)
        self.assertIs(spec, self.attributes[1].parent)
        self.assertIs(spec, self.datasets[0].parent)
        self.assertIs(spec, self.datasets[1].parent)
        json.dumps(spec)

    def test_constructor_datatype(self):
        spec = GroupSpec('A test group',
                         name='root_constructor_datatype',
                         datasets=self.datasets,
                         attributes=self.attributes,
                         linkable=False,
                         data_type_def='EphysData')
        self.assertFalse(spec['linkable'])
        self.assertListEqual(spec['attributes'], self.attributes)
        self.assertListEqual(spec['datasets'], self.datasets)
        self.assertEqual(spec['data_type_def'], 'EphysData')
        self.assertIs(spec, self.attributes[0].parent)
        self.assertIs(spec, self.attributes[1].parent)
        self.assertIs(spec, self.datasets[0].parent)
        self.assertIs(spec, self.datasets[1].parent)
        self.assertEqual(spec.data_type_def, 'EphysData')
        self.assertIsNone(spec.data_type_inc)
        json.dumps(spec)

    def test_set_dataset(self):
        spec = GroupSpec('A test group',
                         name='root_test_set_dataset',
                         linkable=False,
                         data_type_def='EphysData')
        spec.set_dataset(self.datasets[0])
        self.assertIs(spec, self.datasets[0].parent)

    def test_set_group(self):
        spec = GroupSpec('A test group',
                         name='root_test_set_group',
                         linkable=False,
                         data_type_def='EphysData')
        spec.set_group(self.subgroups[0])
        spec.set_group(self.subgroups[1])
        self.assertListEqual(spec['groups'], self.subgroups)
        self.assertIs(spec, self.subgroups[0].parent)
        self.assertIs(spec, self.subgroups[1].parent)
        json.dumps(spec)

    def test_type_extension(self):
        spec = GroupSpec('A test group',
                         name='parent_type',
                         datasets=self.datasets,
                         attributes=self.attributes,
                         linkable=False,
                         data_type_def='EphysData')
        dset1_attributes_ext = [
            AttributeSpec('dset1_extra_attribute', 'an extra attribute for the first dataset', 'text')
        ]
        ext_datasets = [
            DatasetSpec('my first dataset extension',
                        'int',
                        name='dataset1',
                        attributes=dset1_attributes_ext,
                        linkable=True),
        ]
        ext_attributes = [
            AttributeSpec('ext_extra_attribute', 'an extra attribute for the group', 'text'),
        ]
        ext = GroupSpec('A test group extension',
                        name='child_type',
                        datasets=ext_datasets,
                        attributes=ext_attributes,
                        linkable=False,
                        data_type_inc=spec,
                        data_type_def='SpikeData')
        ext_dset1 = ext.get_dataset('dataset1')
        ext_dset1_attrs = ext_dset1.attributes
        self.assertDictEqual(ext_dset1_attrs[0], dset1_attributes_ext[0])
        self.assertDictEqual(ext_dset1_attrs[1], self.dset1_attributes[0])
        self.assertDictEqual(ext_dset1_attrs[2], self.dset1_attributes[1])
        self.assertEqual(ext.data_type_def, 'SpikeData')
        self.assertEqual(ext.data_type_inc, 'EphysData')

        ext_dset2 = ext.get_dataset('dataset2')
        self.maxDiff = None
        # this will suffice for now,  assertDictEqual doesn't do deep equality checks
        self.assertEqual(str(ext_dset2), str(self.datasets[1]))
        self.assertAttributesEqual(ext_dset2, self.datasets[1])

        # self.ns_attr_spec
        ndt_attr_spec = AttributeSpec('data_type', 'the data type of this object',  # noqa: F841
                                      'text', value='SpikeData')

        res_attrs = ext.attributes
        self.assertDictEqual(res_attrs[0], ext_attributes[0])
        self.assertDictEqual(res_attrs[1], self.attributes[0])
        self.assertDictEqual(res_attrs[2], self.attributes[1])

        # test that inherited specs are tracked appropriate
        for d in self.datasets:
            with self.subTest(dataset=d.name):
                self.assertTrue(ext.is_inherited_spec(d))
                self.assertFalse(spec.is_inherited_spec(d))

        json.dumps(spec)

    def assertDatasetsEqual(self, spec1, spec2):
        spec1_dsets = spec1.datasets
        spec2_dsets = spec2.datasets
        if len(spec1_dsets) != len(spec2_dsets):
            raise AssertionError('different number of AttributeSpecs')
        else:
            for i in range(len(spec1_dsets)):
                self.assertAttributesEqual(spec1_dsets[i], spec2_dsets[i])

    def assertAttributesEqual(self, spec1, spec2):
        spec1_attr = spec1.attributes
        spec2_attr = spec2.attributes
        if len(spec1_attr) != len(spec2_attr):
            raise AssertionError('different number of AttributeSpecs')
        else:
            for i in range(len(spec1_attr)):
                self.assertDictEqual(spec1_attr[i], spec2_attr[i])

    def test_add_attribute(self):
        spec = GroupSpec('A test group',
                         name='root_constructor',
                         groups=self.subgroups,
                         datasets=self.datasets,
                         linkable=False)
        for attrspec in self.attributes:
            spec.add_attribute(**attrspec)
        self.assertListEqual(spec['attributes'], self.attributes)
        self.assertListEqual(spec['datasets'], self.datasets)
        self.assertNotIn('data_type_def', spec)
        self.assertIs(spec, self.subgroups[0].parent)
        self.assertIs(spec, self.subgroups[1].parent)
        self.assertIs(spec, spec.attributes[0].parent)
        self.assertIs(spec, spec.attributes[1].parent)
        self.assertIs(spec, self.datasets[0].parent)
        self.assertIs(spec, self.datasets[1].parent)
        json.dumps(spec)

    def test_update_attribute_spec(self):
        spec = GroupSpec('A test group',
                         name='root_constructor',
                         attributes=[AttributeSpec('attribute1', 'my first attribute', 'text'), ])
        spec.set_attribute(AttributeSpec('attribute1', 'my first attribute', 'int', value=5))
        res = spec.get_attribute('attribute1')
        self.assertEqual(res.value, 5)
        self.assertEqual(res.dtype, 'int')



import json

from hdmf.spec import AttributeSpec
from hdmf.testing import TestCase


class AttributeSpecTests(TestCase):

    def test_constructor(self):
        spec = AttributeSpec('attribute1',
                             'my first attribute',
                             'text')
        self.assertEqual(spec['name'], 'attribute1')
        self.assertEqual(spec['dtype'], 'text')
        self.assertEqual(spec['doc'], 'my first attribute')
        self.assertIsNone(spec.parent)
        json.dumps(spec)  # to ensure there are no circular links

    def test_invalid_dtype(self):
        with self.assertRaises(ValueError):
            AttributeSpec(name='attribute1',
                          doc='my first attribute',
                          dtype='invalid'              # <-- Invalid dtype must raise a ValueError
                          )

    def test_both_value_and_default_value_set(self):
        with self.assertRaises(ValueError):
            AttributeSpec(name='attribute1',
                          doc='my first attribute',
                          dtype='int',
                          value=5,
                          default_value=10            # <-- Default_value and value can't be set at the same time
                          )

    def test_colliding_shape_and_dims(self):
        with self.assertRaises(ValueError):
            AttributeSpec(name='attribute1',
                          doc='my first attribute',
                          dtype='int',
                          dims=['test'],
                          shape=[None, 2]           # <-- Length of shape and dims do not match must raise a ValueError
                          )

    def test_default_value(self):
        spec = AttributeSpec('attribute1',
                             'my first attribute',
                             'text',
                             default_value='some text')
        self.assertEqual(spec['default_value'], 'some text')
        self.assertEqual(spec.default_value, 'some text')

    def test_shape(self):
        shape = [None, 2]
        spec = AttributeSpec('attribute1',
                             'my first attribute',
                             'text',
                             shape=shape)
        self.assertEqual(spec['shape'], shape)
        self.assertEqual(spec.shape, shape)

    def test_dims_without_shape(self):
        spec = AttributeSpec('attribute1',
                             'my first attribute',
                             'text',
                             dims=['test'])
        self.assertEqual(spec.shape, (None, ))



from hdmf.build import GroupBuilder, DatasetBuilder, LinkBuilder
from hdmf.testing import TestCase


class GroupBuilderSetterTests(TestCase):
    """Tests for setter functions in GroupBuilder class"""

    def setUp(self):
        self.gb = GroupBuilder('gb')
        self.gb2 = GroupBuilder('gb2', source='file1')

    def tearDown(self):
        pass

    def test_setitem_disabled(self):
        """Test __set_item__ is disabled"""
        with self.assertRaises(NotImplementedError):
            self.gb['key'] = 'value'

    def test_add_dataset(self):
        ds = self.gb.add_dataset('my_dataset', list(range(10)))
        self.assertIsInstance(ds, DatasetBuilder)
        self.assertIs(self.gb, ds.parent)

    def test_add_group(self):
        gp = self.gb.add_group('my_subgroup')
        self.assertIsInstance(gp, GroupBuilder)
        self.assertIs(self.gb['my_subgroup'], gp)
        self.assertIs(self.gb, gp.parent)

    def test_add_link(self):
        gp = self.gb.add_group('my_subgroup')
        sl = self.gb.add_link(gp, 'my_link')
        self.assertIsInstance(sl, LinkBuilder)
        self.assertIs(self.gb['my_link'], sl)
        self.assertIs(self.gb, sl.parent)

    def test_add_external_link(self):
        gp = self.gb2.add_group('my_subgroup')
        el = self.gb.add_link(gp, 'my_externallink')
        self.assertIsInstance(el, LinkBuilder)
        self.assertIs(self.gb['my_externallink'], el)
        self.assertIs(self.gb, el.parent)
        self.assertIs(self.gb2, gp.parent)

    # @unittest.expectedFailure
    def test_set_attribute(self):
        self.gb.set_attribute('key', 'value')
        self.assertIn('key', self.gb.obj_type)
        # self.assertEqual(dict.__getitem__(self.gb, 'attributes')['key'], 'value')
        self.assertEqual(self.gb['key'], 'value')

    def test_parent_constructor(self):
        gb2 = GroupBuilder('gb2', parent=self.gb)
        self.assertIs(gb2.parent, self.gb)

    def test_set_group(self):
        self.gb.set_group(self.gb2)
        self.assertIs(self.gb2.parent, self.gb)


class GroupBuilderGetterTests(TestCase):

    def setUp(self):
        self.subgroup1 = GroupBuilder('subgroup1')
        self.dataset1 = DatasetBuilder('dataset1', list(range(10)))
        self.soft_link1 = LinkBuilder(self.subgroup1, 'soft_link1')
        self.int_attr = 1
        self.str_attr = "my_str"

        self.group1 = GroupBuilder('group1', {'subgroup1': self.subgroup1})
        self.gb = GroupBuilder('gb', {'group1': self.group1},
                               {'dataset1': self.dataset1},
                               {'int_attr': self.int_attr,
                                'str_attr': self.str_attr},
                               {'soft_link1': self.soft_link1})
        # {'soft_link1': self.soft_link1,
        #  'external_link1': self.external_link1}))

    def tearDown(self):
        pass

    def test_path(self):
        self.assertEqual(self.subgroup1.path, 'gb/group1/subgroup1')
        self.assertEqual(self.dataset1.path, 'gb/dataset1')
        self.assertEqual(self.soft_link1.path, 'gb/soft_link1')
        self.assertEqual(self.group1.path, 'gb/group1')
        self.assertEqual(self.gb.path, 'gb')

    def test_get_item_group(self):
        """Test __get_item__ for groups"""
        self.assertIs(self.gb['group1'], self.group1)

    def test_get_item_group_subgroup1(self):
        """Test __get_item__ for groups deeper in hierarchy"""
        self.assertIs(self.gb['group1/subgroup1'], self.subgroup1)

    def test_get_item_dataset(self):
        """Test __get_item__ for datasets"""
        self.assertIs(self.gb['dataset1'], self.dataset1)

    def test_get_item_attr1(self):
        """Test __get_item__ for attributes"""
        self.assertEqual(self.gb['int_attr'], self.int_attr)

    def test_get_item_attr2(self):
        """Test __get_item__ for attributes"""
        self.assertEqual(self.gb['str_attr'], self.str_attr)

    def test_get_item_invalid_key(self):
        """Test __get_item__ for invalid key"""
        with self.assertRaises(KeyError):
            self.gb['invalid_key']

    def test_get_item_soft_link(self):
        """Test __get_item__ for soft links"""
        self.assertIs(self.gb['soft_link1'], self.soft_link1)

    def test_get_group(self):
        """Test get() for groups"""
        self.assertIs(self.gb.get('group1'), self.group1)

    def test_get_group_subgroup1(self):
        """Test get() for groups deeper in hierarchy"""
        self.assertIs(self.gb.get('group1/subgroup1'), self.subgroup1)

    def test_get_dataset(self):
        """Test get() for datasets"""
        self.assertIs(self.gb.get('dataset1'), self.dataset1)

    def test_get_attr1(self):
        """Test get() for attributes"""
        self.assertEqual(self.gb.get('int_attr'), self.int_attr)

    def test_get_attr2(self):
        """Test get() for attributes"""
        self.assertEqual(self.gb.get('str_attr'), self.str_attr)

    def test_get_item_fcn_soft_link(self):
        """Test get() for soft links"""
        self.assertIs(self.gb.get('soft_link1'), self.soft_link1)

    def test_get_invalid_key(self):
        """Test get() for invalid key"""
        self.assertIs(self.gb.get('invalid_key'), None)

    def test_items(self):
        """Test items()"""
        items = (
            ('group1', self.group1),
            ('dataset1', self.dataset1),
            ('int_attr', self.int_attr),
            ('str_attr', self.str_attr),
            ('soft_link1', self.soft_link1),
            # ('external_link1', self.external_link1)
        )
        # self.assertSetEqual(items, set(self.gb.items()))
        try:
            self.assertCountEqual(items, self.gb.items())
        except AttributeError:
            self.assertItemsEqual(items, self.gb.items())

    def test_keys(self):
        """Test keys()"""
        keys = (
            'group1',
            'dataset1',
            'int_attr',
            'str_attr',
            'soft_link1',
            # 'external_link1',
        )
        try:
            self.assertCountEqual(keys, self.gb.keys())
        except AttributeError:
            self.assertItemsEqual(keys, self.gb.keys())

    def test_values(self):
        """Test values()"""
        values = (
            self.group1,
            self.dataset1,
            self.int_attr, self.str_attr,
            self.soft_link1,
            # self.external_link1,
        )
        try:
            self.assertCountEqual(values, self.gb.values())
        except AttributeError:
            self.assertItemsEqual(values, self.gb.values())


class GroupBuilderIsEmptyTests(TestCase):

    def test_is_empty_true(self):
        """Test empty when group has nothing in it"""
        gb = GroupBuilder('gb')
        self.assertEqual(gb.is_empty(), True)

    def test_is_empty_true_group(self):
        """Test is_empty() when group has an empty subgroup"""
        gb = GroupBuilder('gb', {'my_subgroup': GroupBuilder('my_subgroup')})
        self.assertEqual(gb.is_empty(), True)

    def test_is_empty_false_dataset(self):
        """Test is_empty() when group has a dataset"""
        gb = GroupBuilder('gb', datasets={'my_dataset': DatasetBuilder('my_dataset')})
        self.assertEqual(gb.is_empty(), False)

    def test_is_empty_false_group_dataset(self):
        """Test is_empty() when group has a subgroup with a dataset"""
        gb = GroupBuilder(
            'gb',
            {'my_subgroup':
             GroupBuilder(
                 'my_subgroup',
                 datasets={'my_dataset': DatasetBuilder('my_dataset')})})
        self.assertEqual(gb.is_empty(), False)

    def test_is_empty_false_attribute(self):
        """Test is_empty() when group has an attribute"""
        gb = GroupBuilder('gb', attributes={'my_attr': 'attr_value'})
        self.assertEqual(gb.is_empty(), False)

    def test_is_empty_false_group_attribute(self):
        """Test is_empty() when group has subgroup with an attribute"""
        gb = GroupBuilder('gb', {'my_subgroup': GroupBuilder('my_subgroup', attributes={'my_attr': 'attr_value'})})
        self.assertEqual(gb.is_empty(), False)


class GroupBuilderDeepUpdateTests(TestCase):

    def test_mutually_exclusive_subgroups(self):
        gb1 = GroupBuilder('gb1', {'subgroup1': GroupBuilder('subgroup1')})
        gb2 = GroupBuilder('gb2', {'subgroup2': GroupBuilder('subgroup2')})
        gb1.deep_update(gb2)
        self.assertIn('subgroup2', gb1)
        gb1sg = gb1['subgroup2']
        gb2sg = gb2['subgroup2']
        self.assertIs(gb1sg, gb2sg)

    def test_mutually_exclusive_datasets(self):
        gb1 = GroupBuilder('gb1', datasets={'dataset1': DatasetBuilder('dataset1', [1, 2, 3])})
        gb2 = GroupBuilder('gb2', datasets={'dataset2': DatasetBuilder('dataset2', [4, 5, 6])})
        gb1.deep_update(gb2)
        self.assertIn('dataset2', gb1)
        # self.assertIs(gb1['dataset2'], gb2['dataset2'])
        self.assertListEqual(gb1['dataset2'].data, gb2['dataset2'].data)

    def test_mutually_exclusive_attributes(self):
        gb1 = GroupBuilder('gb1', attributes={'attr1': 'my_attribute1'})
        gb2 = GroupBuilder('gb2', attributes={'attr2': 'my_attribute2'})
        gb1.deep_update(gb2)
        self.assertIn('attr2', gb2)
        self.assertEqual(gb2['attr2'], 'my_attribute2')

    def test_mutually_exclusive_links(self):
        gb1 = GroupBuilder('gb1', links={'link1': LinkBuilder(GroupBuilder('target1'), 'link1')})
        gb2 = GroupBuilder('gb2', links={'link2': LinkBuilder(GroupBuilder('target2'), 'link2')})
        gb1.deep_update(gb2)
        self.assertIn('link2', gb2)
        self.assertEqual(gb1['link2'], gb2['link2'])

    def test_intersecting_subgroups(self):
        subgroup2 = GroupBuilder('subgroup2')
        gb1 = GroupBuilder('gb1', {'subgroup1': GroupBuilder('subgroup1'), 'subgroup2': subgroup2})
        gb2 = GroupBuilder('gb2', {'subgroup2': GroupBuilder('subgroup2'), 'subgroup3': GroupBuilder('subgroup3')})
        gb1.deep_update(gb2)
        self.assertIn('subgroup3', gb1)
        self.assertIs(gb1['subgroup3'], gb2['subgroup3'])
        self.assertIs(gb1['subgroup2'], subgroup2)

    def test_intersecting_datasets(self):
        gb1 = GroupBuilder('gb1', datasets={'dataset2': DatasetBuilder('dataset2', [1, 2, 3])})
        gb2 = GroupBuilder('gb2', datasets={'dataset2': DatasetBuilder('dataset2', [4, 5, 6])})
        gb1.deep_update(gb2)
        self.assertIn('dataset2', gb1)
        self.assertListEqual(gb1['dataset2'].data, gb2['dataset2'].data)

    def test_intersecting_attributes(self):
        gb1 = GroupBuilder('gb1', attributes={'attr2': 'my_attribute1'})
        gb2 = GroupBuilder('gb2', attributes={'attr2': 'my_attribute2'})
        gb1.deep_update(gb2)
        self.assertIn('attr2', gb2)
        self.assertEqual(gb2['attr2'], 'my_attribute2')

    def test_intersecting_links(self):
        gb1 = GroupBuilder('gb1', links={'link2': LinkBuilder(GroupBuilder('target1'), 'link2')})
        gb2 = GroupBuilder('gb2', links={'link2': LinkBuilder(GroupBuilder('target2'), 'link2')})
        gb1.deep_update(gb2)
        self.assertIn('link2', gb2)
        self.assertEqual(gb1['link2'], gb2['link2'])


class DatasetBuilderDeepUpdateTests(TestCase):

    def test_overwrite(self):
        db1 = DatasetBuilder('db1', [1, 2, 3])
        db2 = DatasetBuilder('db2', [4, 5, 6])
        db1.deep_update(db2)
        self.assertListEqual(db1.data, db2.data)

    def test_no_overwrite(self):
        db1 = DatasetBuilder('db1', [1, 2, 3])
        db2 = DatasetBuilder('db2', [4, 5, 6], attributes={'attr1': 'va1'})
        db1.deep_update(db2)
        self.assertListEqual(db1.data, db2.data)
        self.assertIn('attr1', db1.attributes)



import os
import datetime

from hdmf.spec.write import NamespaceBuilder, YAMLSpecWriter, export_spec
from hdmf.spec.namespace import SpecNamespace, NamespaceCatalog
from hdmf.spec.spec import GroupSpec
from hdmf.testing import TestCase


class TestSpec(TestCase):

    def setUp(self):
        # create a builder for the namespace
        self.ns_name = "mylab"
        self.date = datetime.datetime.now()

        self.ns_builder = NamespaceBuilder(doc="mydoc",
                                           name=self.ns_name,
                                           full_name="My Laboratory",
                                           version="0.0.1",
                                           author="foo",
                                           contact="foo@bar.com",
                                           namespace_cls=SpecNamespace,
                                           date=self.date)

        # create extensions
        ext1 = GroupSpec('A custom DataSeries interface',
                         attributes=[],
                         datasets=[],
                         groups=[],
                         data_type_inc=None,
                         data_type_def='MyDataSeries')

        ext2 = GroupSpec('An extension of a DataSeries interface',
                         attributes=[],
                         datasets=[],
                         groups=[],
                         data_type_inc='MyDataSeries',
                         data_type_def='MyExtendedMyDataSeries')

        ext2.add_dataset(doc='test',
                         dtype='float',
                         name='testdata')

        self.data_types = [ext1, ext2]

        # add the extension
        self.ext_source_path = 'mylab.extensions.yaml'
        self.namespace_path = 'mylab.namespace.yaml'

    def _test_extensions_file(self):
        with open(self.ext_source_path, 'r') as file:
            match_str = \
"""groups:
- data_type_def: MyDataSeries
  doc: A custom DataSeries interface
- data_type_def: MyExtendedMyDataSeries
  data_type_inc: MyDataSeries
  doc: An extension of a DataSeries interface
  datasets:
  - name: testdata
    dtype: float
    doc: test
"""  # noqa: E128
            nsstr = file.read()
            self.assertEqual(nsstr, match_str)

    def _test_namespace_file(self):
        with open(self.namespace_path, 'r') as file:
            match_str = \
"""namespaces:
- author: foo
  contact: foo@bar.com
  date: '%s'
  doc: mydoc
  full_name: My Laboratory
  name: mylab
  schema:
  - doc: Extensions for my lab
    source: mylab.extensions.yaml
    title: Extensions for my lab
  version: 0.0.1
""" % self.date.isoformat()  # noqa: E128
            nsstr = file.read()
            self.assertEqual(nsstr, match_str)


class TestNamespaceBuilder(TestSpec):
    NS_NAME = 'test_ns'

    def setUp(self):
        super().setUp()
        for data_type in self.data_types:
            self.ns_builder.add_spec(source=self.ext_source_path, spec=data_type)
        self.ns_builder.add_source(source=self.ext_source_path,
                                   doc='Extensions for my lab',
                                   title='My lab extensions')
        self.ns_builder.export(self.namespace_path)

    def tearDown(self):
        if os.path.exists(self.ext_source_path):
            os.remove(self.ext_source_path)
        if os.path.exists(self.namespace_path):
            os.remove(self.namespace_path)

    def test_export_namespace(self):
        self._test_namespace_file()
        self._test_extensions_file()

    def test_read_namespace(self):
        ns_catalog = NamespaceCatalog()
        ns_catalog.load_namespaces(self.namespace_path, resolve=True)
        loaded_ns = ns_catalog.get_namespace(self.ns_name)
        self.assertEqual(loaded_ns.doc, "mydoc")
        self.assertEqual(loaded_ns.author, "foo")
        self.assertEqual(loaded_ns.contact, "foo@bar.com")
        self.assertEqual(loaded_ns.full_name, "My Laboratory")
        self.assertEqual(loaded_ns.name, "mylab")
        self.assertEqual(loaded_ns.date, self.date.isoformat())
        self.assertDictEqual(loaded_ns.schema[0], {'doc': 'Extensions for my lab',
                                                   'source': 'mylab.extensions.yaml',
                                                   'title': 'Extensions for my lab'})
        self.assertEqual(loaded_ns.version, "0.0.1")

    def test_get_source_files(self):
        ns_catalog = NamespaceCatalog()
        ns_catalog.load_namespaces(self.namespace_path, resolve=True)
        loaded_ns = ns_catalog.get_namespace(self.ns_name)
        self.assertListEqual(loaded_ns.get_source_files(), ['mylab.extensions.yaml'])

    def test_get_source_description(self):
        ns_catalog = NamespaceCatalog()
        ns_catalog.load_namespaces(self.namespace_path, resolve=True)
        loaded_ns = ns_catalog.get_namespace(self.ns_name)
        descr = loaded_ns.get_source_description('mylab.extensions.yaml')
        self.assertDictEqual(descr, {'doc': 'Extensions for my lab',
                                     'source': 'mylab.extensions.yaml',
                                     'title': 'Extensions for my lab'})

    def test_missing_version(self):
        """Test that creating a namespace builder without a version raises an error."""
        msg = "Namespace '%s' missing key 'version'. Please specify a version for the extension." % self.ns_name
        with self.assertRaisesWith(ValueError, msg):
            self.ns_builder = NamespaceBuilder(doc="mydoc",
                                               name=self.ns_name,
                                               full_name="My Laboratory",
                                               author="foo",
                                               contact="foo@bar.com",
                                               namespace_cls=SpecNamespace,
                                               date=self.date)


class TestYAMLSpecWrite(TestSpec):

    def setUp(self):
        super().setUp()
        for data_type in self.data_types:
            self.ns_builder.add_spec(source=self.ext_source_path, spec=data_type)
        self.ns_builder.add_source(source=self.ext_source_path,
                                   doc='Extensions for my lab',
                                   title='My lab extensions')

    def tearDown(self):
        if os.path.exists(self.ext_source_path):
            os.remove(self.ext_source_path)
        if os.path.exists(self.namespace_path):
            os.remove(self.namespace_path)

    def test_init(self):
        temp = YAMLSpecWriter('.')
        self.assertEqual(temp._YAMLSpecWriter__outdir, '.')

    def test_write_namespace(self):
        temp = YAMLSpecWriter()
        self.ns_builder.export(self.namespace_path, writer=temp)
        self._test_namespace_file()
        self._test_extensions_file()

    def test_get_name(self):
        self.assertEqual(self.ns_name, self.ns_builder.name)


class TestExportSpec(TestSpec):

    def test_export(self):
        """Test that export_spec writes the correct files."""
        export_spec(self.ns_builder, self.data_types, '.')
        self._test_namespace_file()
        self._test_extensions_file()

    def tearDown(self):
        if os.path.exists(self.ext_source_path):
            os.remove(self.ext_source_path)
        if os.path.exists(self.namespace_path):
            os.remove(self.namespace_path)

    def _test_namespace_file(self):
        with open(self.namespace_path, 'r') as file:
            match_str = \
"""namespaces:
- author: foo
  contact: foo@bar.com
  date: '%s'
  doc: mydoc
  full_name: My Laboratory
  name: mylab
  schema:
  - source: mylab.extensions.yaml
  version: 0.0.1
""" % self.date.isoformat()  # noqa: E128
            nsstr = file.read()
            self.assertEqual(nsstr, match_str)

    def test_missing_data_types(self):
        """Test that calling export_spec on a namespace builder without data types raises a warning."""
        with self.assertWarnsWith(UserWarning, 'No data types specified. Exiting.'):
            export_spec(self.ns_builder, [], '.')



from abc import ABCMeta, abstractmethod
from collections.abc import Iterable
import numpy as np
from warnings import warn
import copy

from .utils import docval, getargs, popargs, docval_macro, get_data_shape


@docval_macro('array_data')
class AbstractDataChunkIterator(metaclass=ABCMeta):
    """
    Abstract iterator class used to iterate over DataChunks.

    Derived classes must ensure that all abstract methods and abstract properties are implemented, in
    particular, dtype, maxshape, __iter__, ___next__, recommended_chunk_shape, and recommended_data_shape.
    """

    @abstractmethod
    def __iter__(self):
        """Return the iterator object"""
        raise NotImplementedError("__iter__ not implemented for derived class")

    @abstractmethod
    def __next__(self):
        r"""
        Return the next data chunk or raise a StopIteration exception if all chunks have been retrieved.

        HINT: numpy.s\_ provides a convenient way to generate index tuples using standard array slicing. This
        is often useful to define the DataChunk.selection of the current chunk

        :returns: DataChunk object with the data and selection of the current chunk
        :rtype: DataChunk
        """
        raise NotImplementedError("__next__ not implemented for derived class")

    @abstractmethod
    def recommended_chunk_shape(self):
        """
        Recommend the chunk shape for the data array.

        :return: NumPy-style shape tuple describing the recommended shape for the chunks of the target
                 array or None. This may or may not be the same as the shape of the chunks returned in the
                 iteration process.
        """
        raise NotImplementedError("recommended_chunk_shape not implemented for derived class")

    @abstractmethod
    def recommended_data_shape(self):
        """
        Recommend the initial shape for the data array.

        This is useful in particular to avoid repeated resized of the target array when reading from
        this data iterator. This should typically be either the final size of the array or the known
        minimal shape of the array.

        :return: NumPy-style shape tuple indicating the recommended initial shape for the target array.
                 This may or may not be the final full shape of the array, i.e., the array is allowed
                 to grow. This should not be None.
        """
        raise NotImplementedError("recommended_data_shape not implemented for derived class")

    @property
    @abstractmethod
    def dtype(self):
        """
        Define the data type of the array

        :return: NumPy style dtype or otherwise compliant dtype string
        """
        raise NotImplementedError("dtype not implemented for derived class")

    @property
    @abstractmethod
    def maxshape(self):
        """
        Property describing the maximum shape of the data array that is being iterated over

        :return: NumPy-style shape tuple indicating the maxiumum dimensions up to which the dataset may be
                 resized. Axes with None are unlimited.
        """
        raise NotImplementedError("maxshape not implemented for derived class")


class DataChunkIterator(AbstractDataChunkIterator):
    """
    Custom iterator class used to iterate over chunks of data.

    This default implementation of AbstractDataChunkIterator accepts any iterable and assumes that we iterate over
    a single dimension of the data array (default: the first dimension). DataChunkIterator supports buffered read,
    i.e., multiple values from the input iterator can be combined to a single chunk. This is
    useful for buffered I/O operations, e.g., to improve performance by accumulating data
    in memory and writing larger blocks at once.
    """

    __docval_init = (
            {'name': 'data', 'type': None, 'doc': 'The data object used for iteration', 'default': None},
            {'name': 'maxshape', 'type': tuple,
             'doc': 'The maximum shape of the full data array. Use None to indicate unlimited dimensions',
             'default': None},
            {'name': 'dtype', 'type': np.dtype, 'doc': 'The Numpy data type for the array', 'default': None},
            {'name': 'buffer_size', 'type': int, 'doc': 'Number of values to be buffered in a chunk', 'default': 1},
            {'name': 'iter_axis', 'type': int, 'doc': 'The dimension to iterate over', 'default': 0}
            )

    @docval(*__docval_init)
    def __init__(self, **kwargs):
        """Initialize the DataChunkIterator.
        If 'data' is an iterator and 'dtype' is not specified, then next is called on the iterator in order to determine
        the dtype of the data.
        """
        # Get the user parameters
        self.data, self.__maxshape, self.__dtype, self.buffer_size, self.iter_axis = getargs('data',
                                                                                             'maxshape',
                                                                                             'dtype',
                                                                                             'buffer_size',
                                                                                             'iter_axis',
                                                                                             kwargs)
        self.chunk_index = 0
        # Create an iterator for the data if possible
        if isinstance(self.data, Iterable):
            if self.iter_axis != 0 and isinstance(self.data, (list, tuple)):
                warn('Iterating over an axis other than the first dimension of list or tuple data '
                     'involves converting the data object to a numpy ndarray, which may incur a computational '
                     'cost.')
                self.data = np.asarray(self.data)
            if isinstance(self.data, np.ndarray):
                # iterate over the given axis by adding a new view on data (iter only works on the first dim)
                self.__data_iter = iter(np.moveaxis(self.data, self.iter_axis, 0))
            else:
                self.__data_iter = iter(self.data)
        else:
            self.__data_iter = None
        self.__next_chunk = DataChunk(None, None)
        self.__next_chunk_start = 0
        self.__first_chunk_shape = None
        # Determine the shape of the data if possible
        if self.__maxshape is None:
            # If the self.data object identifies its shape, then use it
            if hasattr(self.data,  "shape"):
                self.__maxshape = self.data.shape
                # Avoid the special case of scalar values by making them into a 1D numpy array
                if len(self.__maxshape) == 0:
                    self.data = np.asarray([self.data, ])
                    self.__maxshape = self.data.shape
                    self.__data_iter = iter(self.data)
            # Try to get an accurate idea of __maxshape for other Python data structures if possible.
            # Don't just callget_shape for a generator as that would potentially trigger loading of all the data
            elif isinstance(self.data, list) or isinstance(self.data, tuple):
                self.__maxshape = get_data_shape(self.data, strict_no_data_load=True)

        # If we have a data iterator and do not know the dtype, then read the first chunk
        if self.__data_iter is not None and self.__dtype is None:
            self._read_next_chunk()

        # Determine the type of the data if possible
        if self.__next_chunk.data is not None:
            self.__dtype = self.__next_chunk.data.dtype
            self.__first_chunk_shape = get_data_shape(self.__next_chunk.data)

        if self.__dtype is None:
            raise Exception('Data type could not be determined. Please specify dtype in DataChunkIterator init.')

    @classmethod
    @docval(*__docval_init)
    def from_iterable(cls, **kwargs):
        return cls(**kwargs)

    def __iter__(self):
        """Return the iterator object"""
        return self

    def _read_next_chunk(self):
        """Read a single chunk from self.__data_iter and store the results in self.__next_chunk

        :returns: self.__next_chunk, i.e., the DataChunk object describing the next chunk
        """
        from h5py import Dataset as H5Dataset
        if isinstance(self.data, H5Dataset):
            start_index = self.chunk_index * self.buffer_size
            stop_index = start_index + self.buffer_size
            iter_data_bounds = self.data.shape[self.iter_axis]
            if start_index >= iter_data_bounds:
                self.__next_chunk = DataChunk(None, None)
            else:
                if stop_index > iter_data_bounds:
                    stop_index = iter_data_bounds

                selection = [slice(None)] * len(self.maxshape)
                selection[self.iter_axis] = slice(start_index, stop_index)
                selection = tuple(selection)
                self.__next_chunk.data = self.data[selection]
                self.__next_chunk.selection = selection
        elif self.__data_iter is not None:
            # the pieces in the buffer - first dimension consists of individual calls to next
            iter_pieces = []
            # offset of where data begins - shift the selection of where to place this chunk by this much
            curr_chunk_offset = 0
            read_next_empty = False
            while len(iter_pieces) < self.buffer_size:
                try:
                    dat = next(self.__data_iter)
                    if dat is None and len(iter_pieces) == 0:
                        # Skip forward in our chunk until we find data
                        curr_chunk_offset += 1
                    elif dat is None and len(iter_pieces) > 0:
                        # Stop iteration if we hit empty data while constructing our block
                        # Buffer may not be full.
                        read_next_empty = True
                        break
                    else:
                        # Add pieces of data to our buffer
                        iter_pieces.append(np.asarray(dat))
                except StopIteration:
                    break

            if len(iter_pieces) == 0:
                self.__next_chunk = DataChunk(None, None)  # signal end of iteration
            else:
                # concatenate all the pieces into the chunk along the iteration axis
                piece_shape = list(get_data_shape(iter_pieces[0]))
                piece_shape.insert(self.iter_axis, 1)  # insert the missing axis
                next_chunk_shape = piece_shape.copy()
                next_chunk_shape[self.iter_axis] *= len(iter_pieces)
                next_chunk_size = next_chunk_shape[self.iter_axis]

                # use the piece dtype because the actual dtype may not have been determined yet
                # NOTE: this could be problematic if a generator returns e.g. floats first and ints later
                self.__next_chunk.data = np.empty(next_chunk_shape, dtype=iter_pieces[0].dtype)
                self.__next_chunk.data = np.stack(iter_pieces, axis=self.iter_axis)

                selection = [slice(None)] * len(self.maxshape)
                selection[self.iter_axis] = slice(self.__next_chunk_start + curr_chunk_offset,
                                                  self.__next_chunk_start + curr_chunk_offset + next_chunk_size)
                self.__next_chunk.selection = tuple(selection)

                # next chunk should start at self.__next_chunk.selection[self.iter_axis].stop
                # but if this chunk stopped because of reading empty data, then this should be adjusted by 1
                self.__next_chunk_start = self.__next_chunk.selection[self.iter_axis].stop
                if read_next_empty:
                    self.__next_chunk_start += 1
        else:
            self.__next_chunk = DataChunk(None, None)

        self.chunk_index += 1
        return self.__next_chunk

    def __next__(self):
        r"""Return the next data chunk or raise a StopIteration exception if all chunks have been retrieved.

        HINT: numpy.s\_ provides a convenient way to generate index tuples using standard array slicing. This
        is often useful to define the DataChunk.selection of the current chunk

        :returns: DataChunk object with the data and selection of the current chunk
        :rtype: DataChunk

        """
        # If we have not already read the next chunk, then read it now
        if self.__next_chunk.data is None:
            self._read_next_chunk()
        # If we do not have any next chunk
        if self.__next_chunk.data is None:
            raise StopIteration
        # If this is the first time we see a chunk then remember the size of the first chunk
        if self.__first_chunk_shape is None:
            self.__first_chunk_shape = self.__next_chunk.data.shape
        # Keep the next chunk we need to return
        curr_chunk = DataChunk(self.__next_chunk.data,
                               self.__next_chunk.selection)
        # Remove the data for the next chunk from our list since we are returning it here.
        # This is to allow the GarbageCollector to remmove the data when it goes out of scope and avoid
        # having 2 full chunks in memory if not necessary
        self.__next_chunk.data = None
        # Return the current next chunk
        return curr_chunk

    next = __next__

    @docval(returns='Tuple with the recommended chunk shape or None if no particular shape is recommended.')
    def recommended_chunk_shape(self):
        """Recommend a chunk shape.

        To optimize iterative write the chunk should be aligned with the common shape of chunks returned by __next__
        or if those chunks are too large, then a well-aligned subset of those chunks. This may also be
        any other value in case one wants to recommend chunk shapes to optimize read rather
        than write. The default implementation returns None, indicating no preferential chunking option."""
        return None

    @docval(returns='Recommended initial shape for the full data. This should be the shape of the full dataset' +
                    'if known beforehand or alternatively the minimum shape of the dataset. Return None if no ' +
                    'recommendation is available')
    def recommended_data_shape(self):
        """Recommend an initial shape of the data. This is useful when progressively writing data and
        we want to recommend an initial size for the dataset"""
        if self.maxshape is not None:
            if np.all([i is not None for i in self.maxshape]):
                return self.maxshape
        return self.__first_chunk_shape

    @property
    def maxshape(self):
        """
        Get a shape tuple describing the maximum shape of the array described by this DataChunkIterator. If an iterator
        is provided and no data has been read yet, then the first chunk will be read (i.e., next will be called on the
        iterator) in order to determine the maxshape.

        :return: Shape tuple. None is used for dimenwions where the maximum shape is not known or unlimited.
        """
        if self.__maxshape is None:
            # If no data has been read from the iterator yet, read the first chunk and use it to determine the maxshape
            if self.__data_iter is not None and self.__next_chunk.data is None:
                self._read_next_chunk()

            # Determine maxshape from self.__next_chunk
            if self.__next_chunk.data is None:
                return None
            data_shape = get_data_shape(self.__next_chunk.data)
            self.__maxshape = list(data_shape)
            try:
                # Size of self.__next_chunk.data along self.iter_axis is not accurate for maxshape because it is just a
                # chunk. So try to set maxshape along the dimension self.iter_axis based on the shape of self.data if
                # possible. Otherwise, use None to represent an unlimited size
                if hasattr(self.data, '__len__') and self.iter_axis == 0:
                    # special case of 1-D array
                    self.__maxshape[0] = len(self.data)
                else:
                    self.__maxshape[self.iter_axis] = self.data.shape[self.iter_axis]
            except AttributeError:  # from self.data.shape
                self.__maxshape[self.iter_axis] = None
            self.__maxshape = tuple(self.__maxshape)

        return self.__maxshape

    @property
    def dtype(self):
        """
        Get the value data type

        :return: np.dtype object describing the datatype
        """
        return self.__dtype


class DataChunk:
    """
    Class used to describe a data chunk. Used in DataChunkIterator.
    """
    @docval({'name': 'data', 'type': np.ndarray,
             'doc': 'Numpy array with the data value(s) of the chunk', 'default': None},
            {'name': 'selection', 'type': None,
             'doc': 'Numpy index tuple describing the location of the chunk', 'default': None})
    def __init__(self, **kwargs):
        self.data, self.selection = getargs('data', 'selection', kwargs)

    def __len__(self):
        """Get the number of values in the data chunk"""
        if self.data is not None:
            return len(self.data)
        else:
            return 0

    def __getattr__(self, attr):
        """Delegate retrival of attributes to the data in self.data"""
        return getattr(self.data, attr)

    def __copy__(self):
        newobj = DataChunk(data=self.data,
                           selection=self.selection)
        return newobj

    def __deepcopy__(self, memo):
        result = DataChunk(data=copy.deepcopy(self.data),
                           selection=copy.deepcopy(self.selection))
        memo[id(self)] = result
        return result

    def astype(self, dtype):
        """Get a new DataChunk with the self.data converted to the given type"""
        return DataChunk(data=self.data.astype(dtype),
                         selection=self.selection)

    @property
    def dtype(self):
        """
        Data type of the values in the chunk

        :returns: np.dtype of the values in the DataChunk
        """
        return self.data.dtype


def assertEqualShape(data1,
                     data2,
                     axes1=None,
                     axes2=None,
                     name1=None,
                     name2=None,
                     ignore_undetermined=True):
    """
    Ensure that the shape of data1 and data2 match along the given dimensions

    :param data1: The first input array
    :type data1: List, Tuple, np.ndarray, DataChunkIterator etc.
    :param data2: The second input array
    :type data2: List, Tuple, np.ndarray, DataChunkIterator etc.
    :param name1: Optional string with the name of data1
    :param name2: Optional string with the name of data2
    :param axes1: The dimensions of data1 that should be matched to the dimensions of data2. Set to None to
                  compare all axes in order.
    :type axes1: int, Tuple of ints, List of ints, or None
    :param axes2: The dimensions of data2 that should be matched to the dimensions of data1. Must have
                  the same length as axes1. Set to None to compare all axes in order.
    :type axes1: int, Tuple of ints, List of ints, or None
    :param ignore_undetermined: Boolean indicating whether non-matching unlimited dimensions should be ignored,
               i.e., if two dimension don't match because we can't determine the shape of either one, then
               should we ignore that case or treat it as no match

    :return: Bool indicating whether the check passed and a string with a message about the matching process
    """
    # Create the base return object
    response = ShapeValidatorResult()
    # Determine the shape of the datasets
    response.shape1 = get_data_shape(data1)
    response.shape2 = get_data_shape(data2)
    # Determine the number of dimensions of the datasets
    num_dims_1 = len(response.shape1) if response.shape1 is not None else None
    num_dims_2 = len(response.shape2) if response.shape2 is not None else None
    # Determine the string names of the datasets
    n1 = name1 if name1 is not None else ("data1 at " + str(hex(id(data1))))
    n2 = name2 if name2 is not None else ("data2 at " + str(hex(id(data2))))
    # Determine the axes we should compare
    response.axes1 = list(range(num_dims_1)) if axes1 is None else ([axes1] if isinstance(axes1, int) else axes1)
    response.axes2 = list(range(num_dims_2)) if axes2 is None else ([axes2] if isinstance(axes2, int) else axes2)
    # Validate the array shape
    # 1) Check the number of dimensions of the arrays
    if (response.axes1 is None and response.axes2 is None) and num_dims_1 != num_dims_2:
        response.result = False
        response.error = 'NUM_DIMS_ERROR'
        response.message = response.SHAPE_ERROR[response.error]
        response.message += " %s is %sD and %s is %sD" % (n1, num_dims_1, n2, num_dims_2)
    # 2) Check that we have the same number of dimensions to compare on both arrays
    elif len(response.axes1) != len(response.axes2):
        response.result = False
        response.error = 'NUM_AXES_ERROR'
        response.message = response.SHAPE_ERROR[response.error]
        response.message += " Cannot compare axes %s with %s" % (str(response.axes1), str(response.axes2))
    # 3) Check that the datasets have sufficient numner of dimensions
    elif np.max(response.axes1) >= num_dims_1 or np.max(response.axes2) >= num_dims_2:
        response.result = False
        response.error = 'AXIS_OUT_OF_BOUNDS'
        response.message = response.SHAPE_ERROR[response.error]
        if np.max(response.axes1) >= num_dims_1:
            response.message += "Insufficient number of dimensions for %s -- Expected %i found %i" % \
                                (n1, np.max(response.axes1)+1, num_dims_1)
        elif np.max(response.axes2) >= num_dims_2:
            response.message += "Insufficient number of dimensions for %s -- Expected %i found %i" % \
                                (n2, np.max(response.axes2)+1, num_dims_2)
    # 4) Compare the length of the dimensions we should validate
    else:
        unmatched = []
        ignored = []
        for ax in zip(response.axes1, response.axes2):
            if response.shape1[ax[0]] != response.shape2[ax[1]]:
                if ignore_undetermined and (response.shape1[ax[0]] is None or response.shape2[ax[1]] is None):
                    ignored.append(ax)
                else:
                    unmatched.append(ax)
        response.unmatched = unmatched
        response.ignored = ignored

        # Check if everything checked out
        if len(response.unmatched) == 0:
            response.result = True
            response.error = None
            response.message = response.SHAPE_ERROR[response.error]
            if len(response.ignored) > 0:
                response.message += " Ignored undetermined axes %s" % str(response.ignored)
        else:
            response.result = False
            response.error = 'AXIS_LEN_ERROR'
            response.message = response.SHAPE_ERROR[response.error]
            response.message += "Axes %s with size %s of %s did not match dimensions %s with sizes %s of %s." % \
                                (str([un[0] for un in response.unmatched]),
                                 str([response.shape1[un[0]] for un in response.unmatched]),
                                 n1,
                                 str([un[1] for un in response.unmatched]),
                                 str([response.shape2[un[1]] for un in response.unmatched]),
                                 n2)
            if len(response.ignored) > 0:
                response.message += " Ignored undetermined axes %s" % str(response.ignored)
    return response


class ShapeValidatorResult:
    """Class for storing results from validating the shape of multi-dimensional arrays.

    This class is used to store results generated by ShapeValidator

    :ivar result: Boolean indicating whether results matched or not
    :type result: bool
    :ivar message: Message indicating the result of the matching procedure
    :type messaage: str, None
    """
    SHAPE_ERROR = {None: 'All required axes matched',
                   'NUM_DIMS_ERROR': 'Unequal number of dimensions.',
                   'NUM_AXES_ERROR': "Unequal number of axes for comparison.",
                   'AXIS_OUT_OF_BOUNDS': "Axis index for comparison out of bounds.",
                   'AXIS_LEN_ERROR': "Unequal length of axes."}
    """
    Dict where the Keys are the type of errors that may have occurred during shape comparison and the
    values are strings with default error messages for the type.
    """

    @docval({'name': 'result', 'type': bool, 'doc': 'Result of the shape validation', 'default': False},
            {'name': 'message', 'type': str,
             'doc': 'Message describing the result of the shape validation', 'default': None},
            {'name': 'ignored', 'type': tuple,
             'doc': 'Axes that have been ignored in the validaton process', 'default': tuple(), 'shape': (None, )},
            {'name': 'unmatched', 'type': tuple,
             'doc': 'List of axes that did not match during shape validation', 'default': tuple(), 'shape': (None, )},
            {'name': 'error', 'type': str, 'doc': 'Error that may have occurred. One of ERROR_TYPE', 'default': None},
            {'name': 'shape1', 'type': tuple,
             'doc': 'Shape of the first array for comparison', 'default': tuple(), 'shape': (None, )},
            {'name': 'shape2', 'type': tuple,
             'doc': 'Shape of the second array for comparison', 'default': tuple(), 'shape': (None, )},
            {'name': 'axes1', 'type': tuple,
             'doc': 'Axes for the first array that should match', 'default': tuple(), 'shape': (None, )},
            {'name': 'axes2', 'type': tuple,
             'doc': 'Axes for the second array that should match', 'default': tuple(), 'shape': (None, )},
            )
    def __init__(self, **kwargs):
        self.result, self.message, self.ignored, self.unmatched, \
            self.error, self.shape1, self.shape2, self.axes1, self.axes2 = getargs(
                'result', 'message', 'ignored', 'unmatched', 'error', 'shape1', 'shape2', 'axes1', 'axes2', kwargs)

    def __setattr__(self, key, value):
        """
        Overwrite to ensure that, e.g., error_message is not set to an illegal value.
        """
        if key == 'error':
            if value not in self.SHAPE_ERROR.keys():
                raise ValueError("Illegal error type. Error must be one of ShapeValidatorResult.SHAPE_ERROR: %s"
                                 % str(self.SHAPE_ERROR))
            else:
                super().__setattr__(key, value)
        elif key in ['shape1', 'shape2', 'axes1', 'axes2', 'ignored', 'unmatched']:  # Make sure we sore tuples
            super().__setattr__(key, tuple(value))
        else:
            super().__setattr__(key, value)

    def __getattr__(self, item):
        """
        Overwrite to allow dynamic retrival of the default message
        """
        if item == 'default_message':
            return self.SHAPE_ERROR[self.error]
        return self.__getattribute__(item)


@docval_macro('data')
class DataIO:
    """
    Base class for wrapping data arrays for I/O. Derived classes of DataIO are typically
    used to pass dataset-specific I/O parameters to the particular HDMFIO backend.
    """
    @docval({'name': 'data', 'type': 'array_data', 'doc': 'the data to be written', 'default': None})
    def __init__(self, **kwargs):
        data = popargs('data', kwargs)
        self.__data = data

    def get_io_params(self):
        """
        Returns a dict with the I/O parameters specified in this DataIO.
        """
        return dict()

    @property
    def data(self):
        """Get the wrapped data object"""
        return self.__data

    @data.setter
    def data(self, val):
        """Set the wrapped data object"""
        if self.__data is not None:
            raise ValueError("cannot overwrite 'data' on DataIO")
        self.__data = val

    def __copy__(self):
        """
        Define a custom copy method for shallow copy..

        This is needed due to delegation of __getattr__ to the data to
        ensure proper copy.

        :return: Shallow copy of self, ie., a new instance of DataIO wrapping the same self.data object
        """
        newobj = DataIO(data=self.data)
        return newobj

    def __deepcopy__(self, memo):
        """
        Define a custom copy method for deep copy.

        This is needed due to delegation of __getattr__ to the data to
        ensure proper copy.

        :param memo:
        :return: Deep copy of self, i.e., a new instance of DataIO wrapping a deepcopy of the
        self.data object.
        """
        result = DataIO(data=copy.deepcopy(self.__data))
        memo[id(self)] = result
        return result

    def __len__(self):
        """Number of values in self.data"""
        if not self.valid:
            raise InvalidDataIOError("Cannot get length of data. Data is not valid.")
        return len(self.data)

    def __bool__(self):
        return self.valid and len(self) > 0

    def __getattr__(self, attr):
        """Delegate attribute lookup to data object"""
        if attr == '__array_struct__' and not self.valid:
            # np.array() checks __array__ or __array_struct__ attribute dep. on numpy version
            raise InvalidDataIOError("Cannot convert data to array. Data is not valid.")
        if not self.valid:
            raise InvalidDataIOError("Cannot get attribute '%s' of data. Data is not valid." % attr)
        return getattr(self.data, attr)

    def __getitem__(self, item):
        """Delegate slicing to the data object"""
        if not self.valid:
            raise InvalidDataIOError("Cannot get item from data. Data is not valid.")
        return self.data[item]

    def __array__(self):
        """
        Support conversion of DataIO.data to a numpy array. This function is
        provided to improve transparent interoperability of DataIO with numpy.

        :return: An array instance of self.data
        """
        if not self.valid:
            raise InvalidDataIOError("Cannot convert data to array. Data is not valid.")
        if hasattr(self.data, '__array__'):
            return self.data.__array__()
        elif isinstance(self.data, DataChunkIterator):
            raise NotImplementedError("Conversion of DataChunkIterator to array not supported")
        else:
            # NOTE this may result in a copy of the array
            return np.asarray(self.data)

    def __next__(self):
        """Delegate iteration interface to data object"""
        if not self.valid:
            raise InvalidDataIOError("Cannot iterate on data. Data is not valid.")
        return self.data.__next__()

    def __iter__(self):
        """Delegate iteration interface to the data object"""
        if not self.valid:
            raise InvalidDataIOError("Cannot iterate on data. Data is not valid.")
        return self.data.__iter__()

    @property
    def valid(self):
        """bool indicating if the data object is valid"""
        return self.data is not None


class InvalidDataIOError(Exception):
    pass



from h5py import Dataset
import numpy as np
import pandas as pd
from collections import OrderedDict

from ..utils import docval, getargs, ExtenderMeta, call_docval_func, popargs, pystr
from ..container import Container, Data

from . import register_class


@register_class('Index')
class Index(Data):

    __fields__ = ("target",)

    @docval({'name': 'name', 'type': str, 'doc': 'the name of this VectorData'},
            {'name': 'data', 'type': ('array_data', 'data'),
             'doc': 'a dataset where the first dimension is a concatenation of multiple vectors'},
            {'name': 'target', 'type': Data,
             'doc': 'the target dataset that this index applies to'})
    def __init__(self, **kwargs):
        call_docval_func(super().__init__, kwargs)


@register_class('VectorData')
class VectorData(Data):
    """
    A n-dimensional dataset representing a column of a DynamicTable.
    If used without an accompanying VectorIndex, first dimension is
    along the rows of the DynamicTable and each step along the first
    dimension is a cell of the larger table. VectorData can also be
    used to represent a ragged array if paired with a VectorIndex.
    This allows for storing arrays of varying length in a single cell
    of the DynamicTable by indexing into this VectorData. The first
    vector is at VectorData[0:VectorIndex(0)+1]. The second vector is at
    VectorData[VectorIndex(0)+1:VectorIndex(1)+1], and so on.
    """

    __fields__ = ("description",)

    @docval({'name': 'name', 'type': str, 'doc': 'the name of this VectorData'},
            {'name': 'description', 'type': str, 'doc': 'a description for this column'},
            {'name': 'data', 'type': ('array_data', 'data'),
             'doc': 'a dataset where the first dimension is a concatenation of multiple vectors', 'default': list()})
    def __init__(self, **kwargs):
        call_docval_func(super().__init__, kwargs)
        self.description = getargs('description', kwargs)

    @docval({'name': 'val', 'type': None, 'doc': 'the value to add to this column'})
    def add_row(self, **kwargs):
        val = getargs('val', kwargs)
        self.append(val)


@register_class('VectorIndex')
class VectorIndex(Index):
    """
    When paired with a VectorData, this allows for storing arrays of varying
    length in a single cell of the DynamicTable by indexing into this VectorData.
    The first vector is at VectorData[0:VectorIndex(0)+1]. The second vector is at
    VectorData[VectorIndex(0)+1:VectorIndex(1)+1], and so on.
    """

    @docval({'name': 'name', 'type': str, 'doc': 'the name of this VectorIndex'},
            {'name': 'data', 'type': ('array_data', 'data'),
             'doc': 'a 1D dataset containing indexes that apply to VectorData object'},
            {'name': 'target', 'type': VectorData,
             'doc': 'the target dataset that this index applies to'})
    def __init__(self, **kwargs):
        call_docval_func(super().__init__, kwargs)
        self.target = getargs('target', kwargs)

    def add_vector(self, arg):
        self.target.extend(arg)
        self.append(len(self.target))

    def add_row(self, arg):
        self.add_vector(arg)

    def __getitem_helper(self, arg):
        start = 0 if arg == 0 else self.data[arg-1]
        end = self.data[arg]
        return self.target[start:end]

    def __getitem__(self, arg):
        if isinstance(arg, slice):
            indices = list(range(*arg.indices(len(self.data))))
            ret = list()
            for i in indices:
                ret.append(self.__getitem_helper(i))
            return ret
        else:
            return self.__getitem_helper(arg)


@register_class('ElementIdentifiers')
class ElementIdentifiers(Data):

    @docval({'name': 'name', 'type': str, 'doc': 'the name of this ElementIdentifiers'},
            {'name': 'data', 'type': ('array_data', 'data'), 'doc': 'a 1D dataset containing identifiers',
             'default': list()})
    def __init__(self, **kwargs):
        call_docval_func(super().__init__, kwargs)

    @docval({'name': 'other', 'type': (Data, np.ndarray, list, tuple, int),
             'doc': 'List of ids to search for in this ElementIdentifer object'},
            rtype=np.ndarray,
            returns='Array with the list of indices where the elements in the list where found.'
                    'Note, the elements in the returned list are ordered in increasing index'
                    'of the found elements, rather than in the order in which the elements'
                    'where given for the search. Also the length of the result may be different from the length'
                    'of the input array. E.g., if our ids are [1,2,3] and we are search for [3,1,5] the '
                    'result would be [0,2] and NOT [2,0,None]')
    def __eq__(self, other):
        """
        Given a list of ids return the indices in the ElementIdentifiers array where the
        indices are found.
        """
        # Determine the ids we want to find
        search_ids = other if not isinstance(other, Data) else other.data
        if isinstance(search_ids, int):
            search_ids = [search_ids]
        # Find all matching locations
        return np.in1d(self.data, search_ids).nonzero()[0]


@register_class('DynamicTable')
class DynamicTable(Container):
    r"""
    A column-based table. Columns are defined by the argument *columns*. This argument
    must be a list/tuple of :class:`~hdmf.common.table.VectorData` and :class:`~hdmf.common.table.VectorIndex` objects
    or a list/tuple of dicts containing the keys ``name`` and ``description`` that provide the name and description
    of each column in the table. Additionally, the keys ``index`` and ``table`` for specifying additional structure to
    the table columns. Setting the key ``index`` to ``True`` can be used to indicate that the
    :class:`~hdmf.common.table.VectorData` column will store a ragged array (i.e. will be accompanied with a
    :class:`~hdmf.common.table.VectorIndex`). Setting the key ``table`` to ``True`` can be used to indicate that the
    column will store regions to another DynamicTable.

    Columns in DynamicTable subclasses can be statically defined by specifying the class attribute *\_\_columns\_\_*,
    rather than specifying them at runtime at the instance level. This is useful for defining a table structure
    that will get reused. The requirements for *\_\_columns\_\_* are the same as the requirements described above
    for specifying table columns with the *columns* argument to the DynamicTable constructor.
    """

    __fields__ = (
        {'name': 'id', 'child': True},
        {'name': 'columns', 'child': True},
        'colnames',
        'description'
    )

    __columns__ = tuple()

    @ExtenderMeta.pre_init
    def __gather_columns(cls, name, bases, classdict):
        '''
        This classmethod will be called during class declaration in the metaclass to automatically
        include all columns declared in subclasses
        '''
        if not isinstance(cls.__columns__, tuple):
            msg = "'__columns__' must be of type tuple, found %s" % type(cls.__columns__)
            raise TypeError(msg)

        if len(bases) and 'DynamicTable' in globals() and issubclass(bases[-1], Container) \
                and bases[-1].__columns__ is not cls.__columns__:
            new_columns = list(cls.__columns__)
            new_columns[0:0] = bases[-1].__columns__
            cls.__columns__ = tuple(new_columns)

    @docval({'name': 'name', 'type': str, 'doc': 'the name of this table'},
            {'name': 'description', 'type': str, 'doc': 'a description of what is in this table'},
            {'name': 'id', 'type': ('array_data', ElementIdentifiers), 'doc': 'the identifiers for this table',
             'default': None},
            {'name': 'columns', 'type': (tuple, list), 'doc': 'the columns in this table', 'default': None},
            {'name': 'colnames', 'type': 'array_data', 'doc': 'the names of the columns in this table',
             'default': None})
    def __init__(self, **kwargs):  # noqa: C901
        id, columns, desc, colnames = popargs('id', 'columns', 'description', 'colnames', kwargs)
        call_docval_func(super().__init__, kwargs)
        self.description = desc

        # All tables must have ElementIdentifiers (i.e. a primary key column)
        # Here, we figure out what to do for that
        if id is not None:
            if not isinstance(id, ElementIdentifiers):
                id = ElementIdentifiers('id', data=id)
        else:
            id = ElementIdentifiers('id')

        if columns is not None:
            if len(columns) > 0:
                # If columns have been passed in, check them over
                # and process accordingly
                if isinstance(columns[0], dict):
                    columns = self.__build_columns(columns)
                elif not all(isinstance(c, (VectorData, VectorIndex)) for c in columns):
                    raise ValueError("'columns' must be a list of VectorData, DynamicTableRegion or VectorIndex")
                colset = {c.name: c for c in columns}
                for c in columns:
                    if isinstance(c, VectorIndex):
                        colset.pop(c.target.name)
                lens = [len(c) for c in colset.values()]
                if not all(i == lens[0] for i in lens):
                    raise ValueError("columns must be the same length")
                if lens[0] != len(id):
                    if len(id) > 0:
                        raise ValueError("must provide same number of ids as length of columns")
                    else:
                        id.data.extend(range(lens[0]))
        else:
            # if the user has not passed in columns, make a place to put them,
            # as they will presumably be adding new columns
            columns = list()

        self.id = id

        if colnames is None:
            if columns is None:
                # make placeholder for columns if nothing was given
                self.colnames = list()
                self.columns = list()
            else:
                # Figure out column names if columns were given
                tmp = list()
                for col in columns:
                    if isinstance(col, VectorIndex):
                        continue
                    tmp.append(col.name)
                self.colnames = tuple(tmp)
                self.columns = columns
        else:
            # Calculate the order of column names
            if columns is None:
                raise ValueError("Must supply 'columns' if specifying 'colnames'")
            else:
                # order the columns according to the column names
                self.colnames = tuple(pystr(c) for c in colnames)
                col_dict = {col.name: col for col in columns}
                order = dict()
                indexed = dict()
                for col in columns:
                    if isinstance(col, VectorIndex):
                        indexed[col.target.name] = True
                    else:
                        if col.name in indexed:
                            continue
                        indexed[col.name] = False
                i = 0
                for name in self.colnames:
                    col = col_dict[name]
                    order[col.name] = i
                    if indexed[col.name]:
                        i = i + 1
                    i = i + 1
                tmp = [None] * i
                for col in columns:
                    if indexed.get(col.name, False):
                        continue
                    if isinstance(col, VectorData):
                        pos = order[col.name]
                        tmp[pos] = col
                    elif isinstance(col, VectorIndex):
                        pos = order[col.target.name]
                        tmp[pos] = col
                        tmp[pos+1] = col.target
                self.columns = list(tmp)

        # to make generating DataFrames and Series easier
        col_dict = dict()
        self.__indices = dict()
        for col in self.columns:
            if hasattr(self, col.name):
                raise ValueError("Column name '%s' is not allowed because it is already an attribute" % col.name)
            setattr(self, col.name, col)
            if isinstance(col, VectorData):
                existing = col_dict.get(col.name)
                # if we added this column using its index, ignore this column
                if existing is not None:
                    if isinstance(existing, VectorIndex):
                        if existing.target.name == col.name:
                            continue
                        else:
                            raise ValueError("duplicate column does not target VectorData '%s'" % col.name)
                    else:
                        raise ValueError("duplicate column found: '%s'" % col.name)
                else:
                    col_dict[col.name] = col
            elif isinstance(col, VectorIndex):
                col_dict[col.target.name] = col  # use target name for reference and VectorIndex for retrieval
                self.__indices[col.name] = col

        self.__df_cols = [self.id] + [col_dict[name] for name in self.colnames]
        self.__colids = {name: i+1 for i, name in enumerate(self.colnames)}
        for col in self.__columns__:
            if col['name'] not in self.__colids:
                if col.get('required', False):
                    self.add_column(col['name'], col['description'],
                                    index=col.get('index', False),
                                    table=col.get('table', False),
                                    # Pass through extra keyword arguments for add_column that subclasses may have added
                                    **{k: col[k] for k in col.keys()
                                       if k not in ['name', 'description', 'index', 'table', 'required']})

                else:  # create column name attributes (set to None) on the object even if column is not required
                    setattr(self, col['name'], None)
                    if col.get('index', False):
                        setattr(self, col['name'] + '_index', None)

    @staticmethod
    def __build_columns(columns, df=None):
        """
        Build column objects according to specifications
        """
        tmp = list()
        for d in columns:
            name = d['name']
            desc = d.get('description', 'no description')
            data = None
            if df is not None:
                data = list(df[name].values)
            if d.get('index', False):
                index_data = None
                if data is not None:
                    index_data = [len(data[0])]
                    for i in range(1, len(data)):
                        index_data.append(len(data[i]) + index_data[i-1])
                    # assume data came in through a DataFrame, so we need
                    # to concatenate it
                    tmp_data = list()
                    for d in data:
                        tmp_data.extend(d)
                    data = tmp_data
                vdata = VectorData(name, desc, data=data)
                vindex = VectorIndex("%s_index" % name, index_data, target=vdata)
                tmp.append(vindex)
                tmp.append(vdata)
            else:
                if data is None:
                    data = list()
                cls = VectorData
                if d.get('table', False):
                    cls = DynamicTableRegion
                tmp.append(cls(name, desc, data=data))
        return tmp

    def __len__(self):
        return len(self.id)

    @docval({'name': 'data', 'type': dict, 'doc': 'the data to put in this row', 'default': None},
            {'name': 'id', 'type': int, 'doc': 'the ID for the row', 'default': None},
            {'name': 'enforce_unique_id', 'type': bool, 'doc': 'enforce that the id in the table must be unique',
             'default': False},
            allow_extra=True)
    def add_row(self, **kwargs):
        '''
        Add a row to the table. If *id* is not provided, it will auto-increment.
        '''
        data, row_id, enforce_unique_id = popargs('data', 'id', 'enforce_unique_id', kwargs)
        data = data if data is not None else kwargs

        extra_columns = set(list(data.keys())) - set(list(self.__colids.keys()))
        missing_columns = set(list(self.__colids.keys())) - set(list(data.keys()))

        # check to see if any of the extra columns just need to be added
        if extra_columns:
            for col in self.__columns__:
                if col['name'] in extra_columns:
                    if data[col['name']] is not None:
                        self.add_column(col['name'], col['description'],
                                        index=col.get('index', False),
                                        table=col.get('table', False),
                                        # Pass through extra keyword arguments for add_column that
                                        # subclasses may have added
                                        **{k: col[k] for k in col.keys()
                                           if k not in ['name', 'description', 'index', 'table', 'required']})
                    extra_columns.remove(col['name'])

        if extra_columns or missing_columns:
            raise ValueError(
                '\n'.join([
                    'row data keys don\'t match available columns',
                    'you supplied {} extra keys: {}'.format(len(extra_columns), extra_columns),
                    'and were missing {} keys: {}'.format(len(missing_columns), missing_columns)
                ])
            )
        if row_id is None:
            row_id = data.pop('id', None)
        if row_id is None:
            row_id = len(self)
        if enforce_unique_id:
            if row_id in self.id:
                raise ValueError("id %i already in the table" % row_id)
        self.id.append(row_id)

        for colname, colnum in self.__colids.items():
            if colname not in data:
                raise ValueError("column '%s' missing" % colname)
            c = self.__df_cols[colnum]
            if isinstance(c, VectorIndex):
                c.add_vector(data[colname])
            else:
                c.add_row(data[colname])

    def __eq__(self, other):
        return self.to_dataframe().equals(other.to_dataframe())

    @docval({'name': 'name', 'type': str, 'doc': 'the name of this VectorData'},
            {'name': 'description', 'type': str, 'doc': 'a description for this column'},
            {'name': 'data', 'type': ('array_data', 'data'),
             'doc': 'a dataset where the first dimension is a concatenation of multiple vectors', 'default': list()},
            {'name': 'table', 'type': (bool, 'DynamicTable'),
             'doc': 'whether or not this is a table region or the table the region applies to', 'default': False},
            {'name': 'index', 'type': (bool, VectorIndex, 'array_data'),
             'doc': 'whether or not this column should be indexed', 'default': False})
    def add_column(self, **kwargs):
        """
        Add a column to this table. If data is provided, it must
        contain the same number of rows as the current state of the table.
        """
        name, data = getargs('name', 'data', kwargs)
        index, table = popargs('index', 'table', kwargs)
        if name in self.__colids:
            msg = "column '%s' already exists in DynamicTable '%s'" % (name, self.name)
            raise ValueError(msg)

        ckwargs = dict(kwargs)
        cls = VectorData

        # Add table if it's been specified
        if table is not False:
            cls = DynamicTableRegion
            if isinstance(table, DynamicTable):
                ckwargs['table'] = table

        col = cls(**ckwargs)
        col.parent = self
        columns = [col]
        setattr(self, name, col)

        # Add index if it's been specified
        if index is not False:
            if isinstance(index, VectorIndex):
                col_index = index
            elif isinstance(index, bool):        # make empty VectorIndex
                if len(col) > 0:
                    raise ValueError("cannot pass empty index with non-empty data to index")
                col_index = VectorIndex(name + "_index", list(), col)
            else:                                # make VectorIndex with supplied data
                if len(col) == 0:
                    raise ValueError("cannot pass non-empty index with empty data to index")
                col_index = VectorIndex(name + "_index", index, col)
            columns.insert(0, col_index)
            if not isinstance(col_index.parent, Container):
                col_index.parent = self
            # else, the ObjectMapper will create a link from self (parent) to col_index (child with existing parent)
            col = col_index
            self.__indices[col_index.name] = col_index
            setattr(self, col_index.name, col_index)

        if len(col) != len(self.id):
            raise ValueError("column must have the same number of rows as 'id'")
        self.__colids[name] = len(self.__df_cols)
        self.fields['colnames'] = tuple(list(self.colnames)+[name])
        self.fields['columns'] = tuple(list(self.columns)+columns)
        self.__df_cols.append(col)

    @docval({'name': 'name', 'type': str, 'doc': 'the name of the DynamicTableRegion object'},
            {'name': 'region', 'type': (slice, list, tuple), 'doc': 'the indices of the table'},
            {'name': 'description', 'type': str, 'doc': 'a brief description of what the region is'})
    def create_region(self, **kwargs):
        region = getargs('region', kwargs)
        if isinstance(region, slice):
            if (region.start is not None and region.start < 0) or (region.stop is not None and region.stop > len(self)):
                msg = 'region slice %s is out of range for this DynamicTable of length ' % (str(region), len(self))
                raise IndexError(msg)
            region = list(range(*region.indices(len(self))))
        else:
            for idx in region:
                if idx < 0 or idx >= len(self):
                    raise IndexError('The index ' + str(idx) +
                                     ' is out of range for this DynamicTable of length '
                                     + str(len(self)))
        desc = getargs('description', kwargs)
        name = getargs('name', kwargs)
        return DynamicTableRegion(name, region, desc, self)

    def __getitem__(self, key):
        ret = None
        if isinstance(key, tuple):
            # index by row and column --> return specific cell
            arg1 = key[0]
            arg2 = key[1]
            if isinstance(arg2, str):
                arg2 = self.__colids[arg2]
            ret = self.__df_cols[arg2][arg1]
        elif isinstance(key, str):
            # index by one string --> return column
            if key in self.__colids:
                ret = self.__df_cols[self.__colids[key]]
            elif key in self.__indices:
                return self.__indices[key]
            else:
                raise KeyError(key)
        else:
            # index by int, list, or slice --> return pandas Dataframe consisting of one or more rows
            # determine the key. If the key is an int, then turn it into a slice to reduce the number of cases below
            arg = key
            if np.issubdtype(type(arg), np.integer):
                arg = np.s_[arg:(arg+1)]
            # index with a python slice (or single integer) to select one or multiple rows
            if isinstance(arg, slice):
                data = OrderedDict()
                for name in self.colnames:
                    col = self.__df_cols[self.__colids[name]]
                    if isinstance(col.data, (Dataset, np.ndarray)) and col.data.ndim > 1:
                        data[name] = [x for x in col[arg]]
                    else:
                        currdata = col[arg]
                        data[name] = currdata
                id_index = self.id.data[arg]
                if np.isscalar(id_index):
                    id_index = [id_index, ]
                ret = pd.DataFrame(data, index=pd.Index(name=self.id.name, data=id_index), columns=self.colnames)
            # index by a list of ints, return multiple rows
            elif isinstance(arg, (tuple, list, np.ndarray)):
                if isinstance(arg, np.ndarray):
                    if len(arg.shape) != 1:
                        raise ValueError("cannot index DynamicTable with multiple dimensions")
                data = OrderedDict()
                for name in self.colnames:
                    col = self.__df_cols[self.__colids[name]]
                    if isinstance(col.data, (Dataset, np.ndarray)) and col.data.ndim > 1:
                        data[name] = [x for x in col[arg]]
                    elif isinstance(col.data, np.ndarray):
                        data[name] = col[arg]
                    else:
                        data[name] = [col[i] for i in arg]
                id_index = (self.id.data[arg]
                            if isinstance(self.id.data, np.ndarray)
                            else [self.id.data[i] for i in arg])
                ret = pd.DataFrame(data, index=pd.Index(name=self.id.name, data=id_index), columns=self.colnames)
            else:
                raise KeyError("Key type not supported by DynamicTable %s" % str(type(arg)))

        return ret

    def __contains__(self, val):
        return val in self.__colids or val in self.__indices

    def get(self, key, default=None):
        if key in self:
            return self[key]
        return default

    @docval({'name': 'exclude', 'type': set, 'doc': ' Set of columns to exclude from the dataframe', 'default': None})
    def to_dataframe(self, **kwargs):
        """
        Produce a pandas DataFrame containing this table's data.
        """
        exclude = popargs('exclude', kwargs)
        if exclude is None:
            exclude = set([])
        data = OrderedDict()
        for name in self.colnames:
            if name in exclude:
                continue
            col = self.__df_cols[self.__colids[name]]

            if isinstance(col.data, (Dataset, np.ndarray)) and col.data.ndim > 1:
                data[name] = [x for x in col[:]]
            else:
                data[name] = col[:]

        return pd.DataFrame(data, index=pd.Index(name=self.id.name, data=self.id.data))

    @classmethod
    @docval(
        {'name': 'df', 'type': pd.DataFrame, 'doc': 'source DataFrame'},
        {'name': 'name', 'type': str, 'doc': 'the name of this table'},
        {
            'name': 'index_column',
            'type': str,
            'doc': 'if provided, this column will become the table\'s index',
            'default': None
        },
        {
            'name': 'table_description',
            'type': str,
            'doc': 'a description of what is in the resulting table',
            'default': ''
        },
        {
            'name': 'columns',
            'type': (list, tuple),
            'doc': 'a list/tuple of dictionaries specifying columns in the table',
            'default': None
        },
        allow_extra=True
    )
    def from_dataframe(cls, **kwargs):
        '''
        Construct an instance of DynamicTable (or a subclass) from a pandas DataFrame.

        The columns of the resulting table are defined by the columns of the
        dataframe and the index by the dataframe's index (make sure it has a
        name!) or by a column whose name is supplied to the index_column
        parameter. We recommend that you supply *columns* - a list/tuple of
        dictionaries containing the name and description of the column- to help
        others understand the contents of your table. See
        :py:class:`~hdmf.common.table.DynamicTable` for more details on *columns*.
        '''

        columns = kwargs.pop('columns')
        df = kwargs.pop('df')
        name = kwargs.pop('name')
        index_column = kwargs.pop('index_column')
        table_description = kwargs.pop('table_description')
        column_descriptions = kwargs.pop('column_descriptions', dict())

        supplied_columns = dict()
        if columns:
            supplied_columns = {x['name']: x for x in columns}

        class_cols = {x['name']: x for x in cls.__columns__}
        required_cols = set(x['name'] for x in cls.__columns__ if 'required' in x and x['required'])
        df_cols = df.columns
        if required_cols - set(df_cols):
            raise ValueError('missing required cols: ' + str(required_cols - set(df_cols)))
        if set(supplied_columns.keys()) - set(df_cols):
            raise ValueError('cols specified but not provided: ' + str(set(supplied_columns.keys()) - set(df_cols)))
        columns = []
        for col_name in df_cols:
            if col_name in class_cols:
                columns.append(class_cols[col_name])
            elif col_name in supplied_columns:
                columns.append(supplied_columns[col_name])
            else:
                columns.append({'name': col_name,
                                'description': column_descriptions.get(col_name, 'no description')})
                if hasattr(df[col_name].iloc[0], '__len__') and not isinstance(df[col_name].iloc[0], str):
                    lengths = [len(x) for x in df[col_name]]
                    if not lengths[1:] == lengths[:-1]:
                        columns[-1].update(index=True)

        if index_column is not None:
            ids = ElementIdentifiers(name=index_column, data=df[index_column].values.tolist())
        else:
            index_name = df.index.name if df.index.name is not None else 'id'
            ids = ElementIdentifiers(name=index_name, data=df.index.values.tolist())

        columns = cls.__build_columns(columns, df=df)

        return cls(name=name, id=ids, columns=columns, description=table_description, **kwargs)

    def copy(self):
        """
        Return a copy of this DynamicTable.
        This is useful for linking.
        """
        kwargs = dict(name=self.name, id=self.id, columns=self.columns, description=self.description,
                      colnames=self.colnames)
        return self.__class__(**kwargs)


@register_class('DynamicTableRegion')
class DynamicTableRegion(VectorData):
    """
    DynamicTableRegion provides a link from one table to an index or region of another. The `table`
    attribute is another `DynamicTable`, indicating which table is referenced. The data is int(s)
    indicating the row(s) (0-indexed) of the target array. `DynamicTableRegion`s can be used to
    associate multiple rows with the same meta-data without data duplication. They can also be used to
    create hierarchical relationships between multiple `DynamicTable`s. `DynamicTableRegion` objects
    may be paired with a `VectorIndex` object to create ragged references, so a single cell of a
    `DynamicTable` can reference many rows of another `DynamicTable`.
    """

    __fields__ = (
        'table',
        'description'
    )

    @docval({'name': 'name', 'type': str, 'doc': 'the name of this VectorData'},
            {'name': 'data', 'type': ('array_data', 'data'),
             'doc': 'a dataset where the first dimension is a concatenation of multiple vectors'},
            {'name': 'description', 'type': str, 'doc': 'a description of what this region represents'},
            {'name': 'table', 'type': DynamicTable,
             'doc': 'the DynamicTable this region applies to', 'default': None})
    def __init__(self, **kwargs):
        t = popargs('table', kwargs)
        call_docval_func(super().__init__, kwargs)
        self.table = t

    @property
    def table(self):
        return self.fields.get('table')

    @table.setter
    def table(self, val):
        if val is None:
            return
        if 'table' in self.fields:
            msg = "can't set attribute 'table' -- already set"
            raise AttributeError(msg)
        for idx in self.data:
            if idx < 0 or idx >= len(val):
                raise IndexError('The index ' + str(idx) +
                                 ' is out of range for this DynamicTable of length '
                                 + str(len(val)))
        self.fields['table'] = val

    def __getitem__(self, key):
        # treat the list of indices as data that can be indexed. then pass the
        # result to the table to get the data
        if isinstance(key, tuple):
            arg1 = key[0]
            arg2 = key[1]
            return self.table[self.data[arg1], arg2]
        elif isinstance(key, (int, slice)):
            if isinstance(key, int) and key >= len(self.data):
                raise IndexError('index {} out of bounds for data of length {}'.format(key, len(self.data)))
            return self.table[self.data[key]]
        else:
            raise ValueError("unrecognized argument: '%s'" % key)

    def to_dataframe(self, **kwargs):
        """
        Convert the whole DynamicTableRegion to a pandas dataframe.

        Keyword arguments are passed through to the to_dataframe method of DynamicTable that
        is being referenced (i.e., self.table). This allows specification of the 'exclude'
        parameter and any other parameters of DynamicTable.to_dataframe.
        """
        return self.table.to_dataframe(**kwargs).iloc[self.data[:]]

    @property
    def shape(self):
        """
        Define the shape, i.e., (num_rows, num_columns) of the selected table region
        :return: Shape tuple with two integers indicating the number of rows and number of columns
        """
        return (len(self.data), len(self.table.columns))

    def __repr__(self):
        cls = self.__class__
        template = "%s %s.%s at 0x%d\n" % (self.name, cls.__module__, cls.__name__, id(self))
        template += "    Target table: %s %s.%s at 0x%d\n" % (self.table.name,
                                                              self.table.__class__.__module__,
                                                              self.table.__class__.__name__,
                                                              id(self.table))
        return template



import numpy as np
from copy import copy, deepcopy

from hdmf.data_utils import DataChunk
from hdmf.testing import TestCase


class DataChunkTests(TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_datachunk_copy(self):
        obj = DataChunk(data=np.arange(3), selection=np.s_[0:3])
        obj_copy = copy(obj)
        self.assertNotEqual(id(obj), id(obj_copy))
        self.assertEqual(id(obj.data), id(obj_copy.data))
        self.assertEqual(id(obj.selection), id(obj_copy.selection))

    def test_datachunk_deepcopy(self):
        obj = DataChunk(data=np.arange(3), selection=np.s_[0:3])
        obj_copy = deepcopy(obj)
        self.assertNotEqual(id(obj), id(obj_copy))
        self.assertNotEqual(id(obj.data), id(obj_copy.data))
        self.assertNotEqual(id(obj.selection), id(obj_copy.selection))

    def test_datachunk_astype(self):
        obj = DataChunk(data=np.arange(3), selection=np.s_[0:3])
        newtype = np.dtype('int16')
        obj_astype = obj.astype(newtype)
        self.assertNotEqual(id(obj), id(obj_astype))
        self.assertEqual(obj_astype.dtype, np.dtype(newtype))



# this prevents breaking of code that imports these classes directly from map.py
from .manager import Proxy, BuildManager, TypeSource, TypeMap  # noqa: F401
from .objectmapper import ObjectMapper  # noqa: F401

import warnings
warnings.warn(PendingDeprecationWarning('Classes in map.py should be imported from hdmf.build. Importing from '
                                        'hdmf.build.map is subject to change in HDMF 2.0.'))



import scipy.sparse as sps
import numpy as np
import h5py

from ..container import Container
from ..utils import docval, getargs, call_docval_func

from . import register_class


@register_class('CSRMatrix')
class CSRMatrix(Container):

    @docval({'name': 'data', 'type': (sps.csr_matrix, np.ndarray, h5py.Dataset),
             'doc': 'the data to use for this CSRMatrix or CSR data array.'
                    'If passing CSR data array, *indices*, *indptr*, and *shape* must also be provided'},
            {'name': 'indices', 'type': (np.ndarray, h5py.Dataset), 'doc': 'CSR index array', 'default': None},
            {'name': 'indptr', 'type': (np.ndarray, h5py.Dataset), 'doc': 'CSR index pointer array', 'default': None},
            {'name': 'shape', 'type': (list, tuple, np.ndarray), 'doc': 'the shape of the matrix', 'default': None},
            {'name': 'name', 'type': str, 'doc': 'the name to use for this when storing', 'default': 'csr_matrix'})
    def __init__(self, **kwargs):
        call_docval_func(super().__init__, kwargs)
        data = getargs('data', kwargs)
        if isinstance(data, (np.ndarray, h5py.Dataset)):
            if data.ndim == 2:
                data = sps.csr_matrix(self.data)
            elif data.ndim == 1:
                indptr, indices, shape = getargs('indptr', 'indices', 'shape', kwargs)
                if any(_ is None for _ in (indptr, indices, shape)):
                    raise ValueError("must specify indptr, indices, and shape when passing data array")
                self.__check_ind(indptr, 'indptr')
                self.__check_ind(indices, 'indices')
                if len(shape) != 2:
                    raise ValueError('shape must specify two and only two dimensions')
                data = sps.csr_matrix((data, indices, indptr), shape=shape)
            else:
                raise ValueError("cannot use ndarray of dimensionality > 2")
        self.__data = data
        self.__shape = data.shape

    @staticmethod
    def __check_ind(ar, arg):
        if not (ar.ndim == 1 or np.issubdtype(ar.dtype, int)):
            raise ValueError('%s must be a 1D array of integers' % arg)

    def __getattr__(self, val):
        return getattr(self.__data, val)

    @property
    def shape(self):
        return self.__shape

    def to_spmat(self):
        return self.__data



from abc import ABCMeta, abstractmethod
from datetime import datetime
from dateutil.tz import tzlocal
import numpy as np

from hdmf.spec import GroupSpec, AttributeSpec, DatasetSpec, SpecCatalog, SpecNamespace
from hdmf.build import GroupBuilder, DatasetBuilder
from hdmf.validate import ValidatorMap
from hdmf.validate.errors import *  # noqa: F403
from hdmf.testing import TestCase

CORE_NAMESPACE = 'test_core'


class ValidatorTestBase(TestCase, metaclass=ABCMeta):

    def setUp(self):
        spec_catalog = SpecCatalog()
        for spec in self.getSpecs():
            spec_catalog.register_spec(spec, 'test.yaml')
        self.namespace = SpecNamespace(
            'a test namespace', CORE_NAMESPACE, [{'source': 'test.yaml'}], version='0.1.0', catalog=spec_catalog)
        self.vmap = ValidatorMap(self.namespace)

    @abstractmethod
    def getSpecs(self):
        pass


class TestEmptySpec(ValidatorTestBase):

    def getSpecs(self):
        return (GroupSpec('A test group specification with a data type', data_type_def='Bar'),)

    def test_valid(self):
        builder = GroupBuilder('my_bar', attributes={'data_type': 'Bar'})
        validator = self.vmap.get_validator('Bar')
        result = validator.validate(builder)
        self.assertEqual(len(result), 0)

    def test_invalid_missing_req_type(self):
        builder = GroupBuilder('my_bar')
        err_msg = r"builder must have data type defined with attribute '[A-Za-z_]+'"
        with self.assertRaisesRegex(ValueError, err_msg):
            self.vmap.validate(builder)


class TestBasicSpec(ValidatorTestBase):

    def getSpecs(self):
        ret = GroupSpec('A test group specification with a data type',
                        data_type_def='Bar',
                        datasets=[DatasetSpec('an example dataset', 'int', name='data',
                                              attributes=[AttributeSpec(
                                                  'attr2', 'an example integer attribute', 'int')])],
                        attributes=[AttributeSpec('attr1', 'an example string attribute', 'text')])
        return (ret,)

    def test_invalid_missing(self):
        builder = GroupBuilder('my_bar', attributes={'data_type': 'Bar'})
        validator = self.vmap.get_validator('Bar')
        result = validator.validate(builder)
        self.assertEqual(len(result), 2)
        self.assertIsInstance(result[0], MissingError)  # noqa: F405
        self.assertEqual(result[0].name, 'Bar/attr1')
        self.assertIsInstance(result[1], MissingError)  # noqa: F405
        self.assertEqual(result[1].name, 'Bar/data')

    def test_invalid_incorrect_type_get_validator(self):
        builder = GroupBuilder('my_bar', attributes={'data_type': 'Bar', 'attr1': 10})
        validator = self.vmap.get_validator('Bar')
        result = validator.validate(builder)
        self.assertEqual(len(result), 2)
        self.assertIsInstance(result[0], DtypeError)  # noqa: F405
        self.assertEqual(result[0].name, 'Bar/attr1')
        self.assertIsInstance(result[1], MissingError)  # noqa: F405
        self.assertEqual(result[1].name, 'Bar/data')

    def test_invalid_incorrect_type_validate(self):
        builder = GroupBuilder('my_bar', attributes={'data_type': 'Bar', 'attr1': 10})
        result = self.vmap.validate(builder)
        self.assertEqual(len(result), 2)
        self.assertIsInstance(result[0], DtypeError)  # noqa: F405
        self.assertEqual(result[0].name, 'Bar/attr1')
        self.assertIsInstance(result[1], MissingError)  # noqa: F405
        self.assertEqual(result[1].name, 'Bar/data')

    def test_valid(self):
        builder = GroupBuilder('my_bar',
                               attributes={'data_type': 'Bar', 'attr1': 'a string attribute'},
                               datasets=[DatasetBuilder('data', 100, attributes={'attr2': 10})])
        validator = self.vmap.get_validator('Bar')
        result = validator.validate(builder)
        self.assertEqual(len(result), 0)


class TestDateTimeInSpec(ValidatorTestBase):

    def getSpecs(self):
        ret = GroupSpec('A test group specification with a data type',
                        data_type_def='Bar',
                        datasets=[DatasetSpec('an example dataset', 'int', name='data',
                                              attributes=[AttributeSpec(
                                                  'attr2', 'an example integer attribute', 'int')]),
                                  DatasetSpec('an example time dataset', 'isodatetime', name='time'),
                                  DatasetSpec('an array of times', 'isodatetime', name='time_array',
                                              dims=('num_times',), shape=(None,))],
                        attributes=[AttributeSpec('attr1', 'an example string attribute', 'text')])
        return (ret,)

    def test_valid_isodatetime(self):
        builder = GroupBuilder('my_bar',
                               attributes={'data_type': 'Bar', 'attr1': 'a string attribute'},
                               datasets=[DatasetBuilder('data', 100, attributes={'attr2': 10}),
                                         DatasetBuilder('time',
                                                        datetime(2017, 5, 1, 12, 0, 0, tzinfo=tzlocal())),
                                         DatasetBuilder('time_array',
                                                        [datetime(2017, 5, 1, 12, 0, 0, tzinfo=tzlocal())])])
        validator = self.vmap.get_validator('Bar')
        result = validator.validate(builder)
        self.assertEqual(len(result), 0)

    def test_invalid_isodatetime(self):
        builder = GroupBuilder('my_bar',
                               attributes={'data_type': 'Bar', 'attr1': 'a string attribute'},
                               datasets=[DatasetBuilder('data', 100, attributes={'attr2': 10}),
                                         DatasetBuilder('time', 100),
                                         DatasetBuilder('time_array',
                                                        [datetime(2017, 5, 1, 12, 0, 0, tzinfo=tzlocal())])])
        validator = self.vmap.get_validator('Bar')
        result = validator.validate(builder)
        self.assertEqual(len(result), 1)
        self.assertIsInstance(result[0], DtypeError)  # noqa: F405
        self.assertEqual(result[0].name, 'Bar/time')

    def test_invalid_isodatetime_array(self):
        builder = GroupBuilder('my_bar',
                               attributes={'data_type': 'Bar', 'attr1': 'a string attribute'},
                               datasets=[DatasetBuilder('data', 100, attributes={'attr2': 10}),
                                         DatasetBuilder('time',
                                                        datetime(2017, 5, 1, 12, 0, 0, tzinfo=tzlocal())),
                                         DatasetBuilder('time_array',
                                                        datetime(2017, 5, 1, 12, 0, 0, tzinfo=tzlocal()))])
        validator = self.vmap.get_validator('Bar')
        result = validator.validate(builder)
        self.assertEqual(len(result), 1)
        self.assertIsInstance(result[0], ExpectedArrayError)  # noqa: F405
        self.assertEqual(result[0].name, 'Bar/time_array')


class TestNestedTypes(ValidatorTestBase):

    def getSpecs(self):
        bar = GroupSpec('A test group specification with a data type',
                        data_type_def='Bar',
                        datasets=[DatasetSpec('an example dataset', 'int', name='data',
                                              attributes=[AttributeSpec('attr2', 'an example integer attribute',
                                                                        'int')])],
                        attributes=[AttributeSpec('attr1', 'an example string attribute', 'text')])
        foo = GroupSpec('A test group that contains a data type',
                        data_type_def='Foo',
                        groups=[GroupSpec('A Bar group for Foos', name='my_bar', data_type_inc='Bar')],
                        attributes=[AttributeSpec('foo_attr', 'a string attribute specified as text', 'text',
                                                  required=False)])

        return (bar, foo)

    def test_invalid_missing_req_group(self):
        foo_builder = GroupBuilder('my_foo', attributes={'data_type': 'Foo',
                                                         'foo_attr': 'example Foo object'})
        results = self.vmap.validate(foo_builder)
        self.assertIsInstance(results[0], MissingDataType)  # noqa: F405
        self.assertEqual(results[0].name, 'Foo')
        self.assertEqual(results[0].reason, 'missing data type Bar')

    def test_invalid_wrong_name_req_type(self):
        bar_builder = GroupBuilder('bad_bar_name',
                                   attributes={'data_type': 'Bar', 'attr1': 'a string attribute'},
                                   datasets=[DatasetBuilder('data', 100, attributes={'attr2': 10})])

        foo_builder = GroupBuilder('my_foo',
                                   attributes={'data_type': 'Foo', 'foo_attr': 'example Foo object'},
                                   groups=[bar_builder])

        results = self.vmap.validate(foo_builder)
        self.assertEqual(len(results), 1)
        self.assertIsInstance(results[0], MissingDataType)  # noqa: F405
        self.assertEqual(results[0].data_type, 'Bar')

    def test_valid(self):
        bar_builder = GroupBuilder('my_bar',
                                   attributes={'data_type': 'Bar', 'attr1': 'a string attribute'},
                                   datasets=[DatasetBuilder('data', 100, attributes={'attr2': 10})])

        foo_builder = GroupBuilder('my_foo',
                                   attributes={'data_type': 'Foo', 'foo_attr': 'example Foo object'},
                                   groups=[bar_builder])

        results = self.vmap.validate(foo_builder)
        self.assertEqual(len(results), 0)

    def test_valid_wo_opt_attr(self):
        bar_builder = GroupBuilder('my_bar',
                                   attributes={'data_type': 'Bar', 'attr1': 'a string attribute'},
                                   datasets=[DatasetBuilder('data', 100, attributes={'attr2': 10})])
        foo_builder = GroupBuilder('my_foo',
                                   attributes={'data_type': 'Foo'},
                                   groups=[bar_builder])

        results = self.vmap.validate(foo_builder)
        self.assertEqual(len(results), 0)


class TestDtypeValidation(TestCase):

    def set_up_spec(self, dtype):
        spec_catalog = SpecCatalog()
        spec = GroupSpec('A test group specification with a data type',
                         data_type_def='Bar',
                         datasets=[DatasetSpec('an example dataset', dtype, name='data')],
                         attributes=[AttributeSpec('attr1', 'an example attribute', dtype)])
        spec_catalog.register_spec(spec, 'test.yaml')
        self.namespace = SpecNamespace(
            'a test namespace', CORE_NAMESPACE, [{'source': 'test.yaml'}], version='0.1.0', catalog=spec_catalog)
        self.vmap = ValidatorMap(self.namespace)

    def test_ascii_for_utf8(self):
        """Test that validator allows ASCII data where UTF8 is specified."""
        self.set_up_spec('text')
        value = b'an ascii string'
        bar_builder = GroupBuilder('my_bar',
                                   attributes={'data_type': 'Bar', 'attr1': value},
                                   datasets=[DatasetBuilder('data', value)])
        results = self.vmap.validate(bar_builder)
        self.assertEqual(len(results), 0)

    def test_utf8_for_ascii(self):
        """Test that validator does not allow UTF8 where ASCII is specified."""
        self.set_up_spec('bytes')
        value = 'a utf8 string'
        bar_builder = GroupBuilder('my_bar',
                                   attributes={'data_type': 'Bar', 'attr1': value},
                                   datasets=[DatasetBuilder('data', value)])
        results = self.vmap.validate(bar_builder)
        result_strings = set([str(s) for s in results])
        expected_errors = {"Bar/attr1 (my_bar.attr1): incorrect type - expected 'bytes', got 'utf'",
                           "Bar/data (my_bar/data): incorrect type - expected 'bytes', got 'utf'"}
        self.assertEqual(result_strings, expected_errors)

    def test_int64_for_int8(self):
        """Test that validator allows int64 data where int8 is specified."""
        self.set_up_spec('int8')
        value = np.int64(1)
        bar_builder = GroupBuilder('my_bar',
                                   attributes={'data_type': 'Bar', 'attr1': value},
                                   datasets=[DatasetBuilder('data', value)])
        results = self.vmap.validate(bar_builder)
        self.assertEqual(len(results), 0)

    def test_int8_for_int64(self):
        """Test that validator does not allow int8 data where int64 is specified."""
        self.set_up_spec('int64')
        value = np.int8(1)
        bar_builder = GroupBuilder('my_bar',
                                   attributes={'data_type': 'Bar', 'attr1': value},
                                   datasets=[DatasetBuilder('data', value)])
        results = self.vmap.validate(bar_builder)
        result_strings = set([str(s) for s in results])
        expected_errors = {"Bar/attr1 (my_bar.attr1): incorrect type - expected 'int64', got 'int8'",
                           "Bar/data (my_bar/data): incorrect type - expected 'int64', got 'int8'"}
        self.assertEqual(result_strings, expected_errors)

    def test_int64_for_numeric(self):
        """Test that validator allows int64 data where numeric is specified."""
        self.set_up_spec('numeric')
        value = np.int64(1)
        bar_builder = GroupBuilder('my_bar',
                                   attributes={'data_type': 'Bar', 'attr1': value},
                                   datasets=[DatasetBuilder('data', value)])
        results = self.vmap.validate(bar_builder)
        self.assertEqual(len(results), 0)

    def test_bool_for_numeric(self):
        """Test that validator does not allow bool data where numeric is specified."""
        self.set_up_spec('numeric')
        value = np.bool(1)
        bar_builder = GroupBuilder('my_bar',
                                   attributes={'data_type': 'Bar', 'attr1': value},
                                   datasets=[DatasetBuilder('data', value)])
        results = self.vmap.validate(bar_builder)
        result_strings = set([str(s) for s in results])
        expected_errors = {"Bar/attr1 (my_bar.attr1): incorrect type - expected 'numeric', got 'bool'",
                           "Bar/data (my_bar/data): incorrect type - expected 'numeric', got 'bool'"}
        self.assertEqual(result_strings, expected_errors)



from hdmf.utils import LabelledDict
from hdmf.testing import TestCase


class MyTestClass:

    def __init__(self, prop1, prop2):
        self._prop1 = prop1
        self._prop2 = prop2

    @property
    def prop1(self):
        return self._prop1

    @property
    def prop2(self):
        return self._prop2


class TestLabelledDict(TestCase):

    def test_constructor(self):
        """Test that constructor sets arguments properly"""
        ld = LabelledDict(label='all_objects', key_attr='prop1')
        self.assertEqual(ld.label, 'all_objects')
        self.assertEqual(ld.key_attr, 'prop1')

    def test_constructor_default(self):
        """Test that constructor sets default key attribute"""
        ld = LabelledDict(label='all_objects')
        self.assertEqual(ld.key_attr, 'name')

    def test_set_key_attr(self):
        """Test that the key attribute cannot be set after initialization"""
        ld = LabelledDict(label='all_objects')
        with self.assertRaisesWith(AttributeError, "can't set attribute"):
            ld.key_attr = 'another_name'

    def test_getitem_unknown_val(self):
        """Test that dict[val] raises an error if there are no matches for val"""
        ld = LabelledDict(label='all_objects', key_attr='prop1')
        with self.assertRaisesWith(KeyError, "'unknown_val'"):
            ld['unknown_val']

    def test_getitem_eqeq_unknown_val(self):
        """Test that dict[key_attr == val] raises an error if there are no matches for val"""
        ld = LabelledDict(label='all_objects', key_attr='prop1')
        with self.assertRaisesWith(KeyError, "'unknown_val'"):
            ld['prop1 == unknown_val']

    def test_getitem_eqeq_other_key_attr(self):
        """Test that dict[key_attr == val] raises an error if there are no matches for other_attr == val"""
        ld = LabelledDict(label='all_objects', key_attr='prop1')
        with self.assertRaisesWith(KeyError, "'unknown_val'"):
            ld['other_attr == unknown_val']

    def test_getitem_eqeq_no_key_attr(self):
        """Test that dict[key_attr == val] raises an error if key_attr is not given"""
        ld = LabelledDict(label='all_objects', key_attr='prop1')
        with self.assertRaisesWith(ValueError, "An attribute name is required before '=='."):
            ld[' == unknown_key']

    def test_getitem_eqeq_no_val(self):
        """Test that dict[key_attr == val] raises an error if val is not given"""
        ld = LabelledDict(label='all_objects', key_attr='prop1')
        with self.assertRaisesWith(ValueError, "A value is required after '=='."):
            ld['prop1 == ']

    def test_getitem_eqeq_no_key_attr_no_val(self):
        """Test that dict[key_attr == val] raises an error if key_attr is not given and val is not given"""
        ld = LabelledDict(label='all_objects', key_attr='prop1')
        with self.assertRaisesWith(ValueError, "An attribute name is required before '=='."):
            ld[' == ']

    def test_add_basic(self):
        """Test add function on object with correct key_attr"""
        ld = LabelledDict(label='all_objects', key_attr='prop1')
        obj1 = MyTestClass('a', 'b')
        ld.add(obj1)
        self.assertIs(ld['a'], obj1)

    def test_add_value_missing_key(self):
        """Test that add raises an error if the value being set does not have the attribute key_attr"""
        ld = LabelledDict(label='all_objects', key_attr='prop3')
        obj1 = MyTestClass('a', 'b')

        err_msg = r"Cannot set value '<.*>' in LabelledDict\. Value must have key 'prop3'\."
        with self.assertRaisesRegex(ValueError, err_msg):
            ld.add(obj1)

    def test_setitem_getitem_basic(self):
        """Test that setitem and getitem properly set and get the object"""
        ld = LabelledDict(label='all_objects', key_attr='prop1')
        obj1 = MyTestClass('a', 'b')
        ld.add(obj1)
        self.assertIs(ld['a'], obj1)

    def test_setitem_value_missing_key(self):
        """Test that setitem raises an error if the value being set does not have the attribute key_attr"""
        ld = LabelledDict(label='all_objects', key_attr='prop3')
        obj1 = MyTestClass('a', 'b')

        err_msg = r"Cannot set value '<.*>' in LabelledDict\. Value must have key 'prop3'\."
        with self.assertRaisesRegex(ValueError, err_msg):
            ld['a'] = obj1

    def test_setitem_value_wrong_value(self):
        """Test that setitem raises an error if the value being set has a different value for attribute key_attr
        than the given key"""
        ld = LabelledDict(label='all_objects', key_attr='prop1')
        obj1 = MyTestClass('a', 'b')

        err_msg = r"Key 'b' must equal attribute 'prop1' of '<.*>'\."
        with self.assertRaisesRegex(KeyError, err_msg):
            ld['b'] = obj1

    def test_addval_getitem_eqeq(self):
        """Test that dict[key_attr == val] returns the single matching object"""
        ld = LabelledDict(label='all_objects', key_attr='prop1')
        obj1 = MyTestClass('a', 'b')
        ld.add(obj1)
        self.assertIs(ld['prop1 == a'], obj1)

    def test_addval_getitem_eqeq_unknown_val(self):
        """Test that dict[key_attr == val] with an unknown val raises an error even if there are other objects in
        dict"""
        ld = LabelledDict(label='all_objects', key_attr='prop1')
        obj1 = MyTestClass('a', 'b')
        ld.add(obj1)
        with self.assertRaisesWith(KeyError, "'unknown_val'"):
            ld['prop1 == unknown_val']

    def test_addval_getitem_eqeq_unknown_key_val(self):
        """Test that dict[key_attr == val] where prop3 does not match any objects in the dict raises an error"""
        ld = LabelledDict(label='all_objects', key_attr='prop1')
        obj1 = MyTestClass('a', 'b')
        ld['a'] = obj1
        with self.assertRaisesWith(KeyError, "'unknown_val'"):
            ld['prop3 == unknown_val']

    def test_addval_getitem_other_key(self):
        """Test that dict[other_key == val] returns a list of matching objects"""
        ld = LabelledDict(label='all_objects', key_attr='prop1')
        obj1 = MyTestClass('a', 'b')
        ld.add(obj1)
        self.assertSetEqual(ld['prop2 == b'], {obj1})

    def test_addval_getitem_other_key_multi(self):
        """Test that dict[other_key == val] returns a list of matching objects"""
        ld = LabelledDict(label='all_objects', key_attr='prop1')
        obj1 = MyTestClass('a', 'b')
        obj2 = MyTestClass('d', 'b')
        obj3 = MyTestClass('f', 'e')
        ld.add(obj1)
        ld.add(obj2)
        ld.add(obj3)
        self.assertSetEqual(ld['prop2 == b'], set([obj1, obj2]))

    def test_addval_getitem_other_key_none(self):
        """Test that dict[other_key == val] raises an error if val does not equal any of the other_key attribute values
        in the dict, even when the other_key attribute exists"""
        ld = LabelledDict(label='all_objects', key_attr='prop1')
        obj1 = MyTestClass('a', 'b')
        obj2 = MyTestClass('d', 'b')
        obj3 = MyTestClass('f', 'e')
        ld.add(obj1)
        ld.add(obj2)
        ld.add(obj3)
        with self.assertRaisesWith(KeyError, "'c'"):
            ld['prop2 == c']



from shutil import copyfile
import os

from hdmf.backends.hdf5.h5tools import HDF5IO
from tests.unit.test_io_hdf5_h5tools import _get_manager
from tests.unit.utils import Foo, FooBucket
from hdmf.testing import TestCase


class Test1_1_0(TestCase):

    def setUp(self):
        # created using manager in test_io_hdf5_h5tools
        self.orig_1_0_5 = 'tests/unit/back_compat_tests/1.0.5.h5'
        self.path_1_0_5 = 'test_1.0.5.h5'
        copyfile(self.orig_1_0_5, self.path_1_0_5)

        # note: this may break if the current manager is different from the old manager
        # better to save a spec file
        self.manager = _get_manager()

    def tearDown(self):
        if os.path.exists(self.path_1_0_5):
            os.remove(self.path_1_0_5)

    def test_read_1_0_5(self):
        '''Test whether we can read files made by hdmf version 1.0.5'''
        with HDF5IO(self.path_1_0_5, manager=self.manager, mode='r') as io:
            read_foofile = io.read()
            self.assertTrue(len(read_foofile.buckets) == 1)
            # workaround for the fact that order of foos is not maintained
            for foo in read_foofile.buckets[0].foos:
                if foo.name == 'foo1':
                    self.assertListEqual(foo.my_data[:].tolist(), [0, 1, 2, 3, 4])
                if foo.name == 'foo2':
                    self.assertListEqual(foo.my_data[:].tolist(), [5, 6, 7, 8, 9])

    def test_append_1_0_5(self):
        '''Test whether we can append to files made by hdmf version 1.0.5'''
        foo = Foo('foo3', [10, 20, 30, 40, 50], "I am foo3", 17, 3.14)
        foobucket = FooBucket('foobucket2', [foo])

        with HDF5IO(self.path_1_0_5, manager=self.manager, mode='a') as io:
            read_foofile = io.read()
            read_foofile.buckets.append(foobucket)
            foobucket.parent = read_foofile
            io.write(read_foofile)

        with HDF5IO(self.path_1_0_5, manager=self.manager, mode='r') as io:
            read_foofile = io.read()
            # workaround for the fact that order of buckets is not maintained
            for bucket in read_foofile.buckets:
                if bucket.name == 'foobucket2':
                    self.assertListEqual(bucket.foos[0].my_data[:].tolist(), foo.my_data)



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

import versioneer

with open('README.rst', 'r') as fp:
    readme = fp.read()

pkgs = find_packages('src', exclude=['data'])
print('found these packages:', pkgs)

schema_dir = 'common/hdmf-common-schema/common'

with open('requirements-min.txt', 'r') as fp:
    # replace == with >= and remove trailing comments and spaces
    reqs = [x.replace('==', '>=').split('#')[0].strip() for x in fp]
    reqs = [x for x in reqs if x]  # remove empty strings

print(reqs)

setup_args = {
    'name': 'hdmf',
    'version': versioneer.get_version(),
    'cmdclass': versioneer.get_cmdclass(),
    'description': 'A package for standardizing hierarchical object data',
    'long_description': readme,
    'long_description_content_type': 'text/x-rst; charset=UTF-8',
    'author': 'Andrew Tritt',
    'author_email': 'ajtritt@lbl.gov',
    'url': 'https://github.com/hdmf-dev/hdmf',
    'license': "BSD",
    'install_requires': reqs,
    'packages': pkgs,
    'package_dir': {'': 'src'},
    'package_data': {'hdmf': ["%s/*.yaml" % schema_dir, "%s/*.json" % schema_dir]},
    'classifiers': [
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: BSD License",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Operating System :: Unix",
        "Topic :: Scientific/Engineering :: Medical Science Apps."
    ],
    'keywords': 'python '
                'HDF '
                'HDF5 '
                'cross-platform '
                'open-data '
                'data-format '
                'open-source '
                'open-science '
                'reproducible-research ',
    'zip_safe': False
}

if __name__ == '__main__':
    setup(**setup_args)



from abc import ABCMeta, abstractmethod

from ..build import BuildManager, GroupBuilder
from ..utils import docval, getargs, popargs
from ..container import Container


class HDMFIO(metaclass=ABCMeta):
    @docval({'name': 'manager', 'type': BuildManager,
             'doc': 'the BuildManager to use for I/O', 'default': None},
            {"name": "source", "type": str,
             "doc": "the source of container being built i.e. file path", 'default': None})
    def __init__(self, **kwargs):
        self.__manager = getargs('manager', kwargs)
        self.__built = dict()
        self.__source = getargs('source', kwargs)
        self.open()

    @property
    def manager(self):
        '''The BuildManager this HDMFIO is using'''
        return self.__manager

    @property
    def source(self):
        '''The source of the container being read/written i.e. file path'''
        return self.__source

    @docval(returns='the Container object that was read in', rtype=Container)
    def read(self, **kwargs):
        f_builder = self.read_builder()
        if all(len(v) == 0 for v in f_builder.values()):
            # TODO also check that the keys are appropriate. print a better error message
            raise UnsupportedOperation('Cannot build data. There are no values.')
        container = self.__manager.construct(f_builder)
        return container

    @docval({'name': 'container', 'type': Container, 'doc': 'the Container object to write'},
            {'name': 'exhaust_dci', 'type': bool,
             'doc': 'exhaust DataChunkIterators one at a time. If False, exhaust them concurrently', 'default': True})
    def write(self, **kwargs):
        container = popargs('container', kwargs)
        f_builder = self.__manager.build(container, source=self.__source)
        self.write_builder(f_builder, **kwargs)

    @abstractmethod
    @docval(returns='a GroupBuilder representing the read data', rtype='GroupBuilder')
    def read_builder(self):
        ''' Read data and return the GroupBuilder representing '''
        pass

    @abstractmethod
    @docval({'name': 'builder', 'type': GroupBuilder, 'doc': 'the GroupBuilder object representing the Container'},
            {'name': 'exhaust_dci', 'type': bool,
             'doc': 'exhaust DataChunkIterators one at a time. If False, exhaust them concurrently', 'default': True})
    def write_builder(self, **kwargs):
        ''' Write a GroupBuilder representing an Container object '''
        pass

    @abstractmethod
    def open(self):
        ''' Open this HDMFIO object for writing of the builder '''
        pass

    @abstractmethod
    def close(self):
        ''' Close this HDMFIO object to further reading/writing'''
        pass

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()


class UnsupportedOperation(ValueError):
    pass



from abc import ABCMeta, abstractmethod

from .utils import docval, getargs, call_docval_func
from .data_utils import AbstractDataChunkIterator, DataChunkIterator, DataChunk


class NotYetExhausted(Exception):
    pass


class DataChunkProcessor(AbstractDataChunkIterator, metaclass=ABCMeta):

    @docval({'name': 'data', 'type': DataChunkIterator, 'doc': 'the DataChunkIterator to analyze'})
    def __init__(self, **kwargs):
        """Initialize the DataChunkIterator"""
        # Get the user parameters
        self.__dci = getargs('data', kwargs)

    def __next__(self):
        try:
            dc = self.__dci.__next__()
        except StopIteration as e:
            self.__done = True
            raise e
        self.process_data_chunk(dc)
        return dc

    def __iter__(self):
        return iter(self.__dci)

    def recommended_chunk_shape(self):
        return self.__dci.recommended_chunk_shape()

    def recommended_data_shape(self):
        return self.__dci.recommended_data_shape()

    def get_final_result(self, **kwargs):
        ''' Return the result of processing data fed by this DataChunkIterator '''
        if not self.__done:
            raise NotYetExhausted()
        return self.compute_final_result()

    @abstractmethod
    @docval({'name': 'data_chunk', 'type': DataChunk, 'doc': 'a chunk to process'})
    def process_data_chunk(self, **kwargs):
        ''' This method should take in a DataChunk,
            and process it.
        '''
        pass

    @abstractmethod
    @docval(returns='the result of processing this stream')
    def compute_final_result(self, **kwargs):
        ''' Return the result of processing this stream
            Should raise NotYetExhaused exception
        '''
        pass


class NumSampleCounter(DataChunkProcessor):

    def __init__(self, **kwargs):
        call_docval_func(super().__init__, kwargs)
        self.__sample_count = 0

    @docval({'name': 'data_chunk', 'type': DataChunk, 'doc': 'a chunk to process'})
    def process_data_chunk(self, **kwargs):
        dc = getargs('data_chunk', kwargs)
        self.__sample_count += len(dc)

    @docval(returns='the result of processing this stream')
    def compute_final_result(self, **kwargs):
        return self.__sample_count



#!/usr/bin/env python

from __future__ import print_function

import warnings
import re
import argparse
import logging
import os.path
import os
import sys
import traceback
import unittest
from tests.coloredtestrunner import ColoredTestRunner, ColoredTestResult

flags = {'hdmf': 1, 'integration': 3, 'example': 4}

TOTAL = 0
FAILURES = 0
ERRORS = 0


def run_test_suite(directory, description="", verbose=True):
    global TOTAL, FAILURES, ERRORS
    logging.info("running %s" % description)
    directory = os.path.join(os.path.dirname(__file__), directory)
    if verbose > 1:
        runner = ColoredTestRunner(verbosity=verbose)
    else:
        runner = unittest.TextTestRunner(verbosity=verbose, resultclass=ColoredTestResult)
    test_result = runner.run(unittest.TestLoader().discover(directory))

    TOTAL += test_result.testsRun
    FAILURES += len(test_result.failures)
    ERRORS += len(test_result.errors)

    return test_result


def _import_from_file(script):
    import imp
    return imp.load_source(os.path.basename(script), script)


warning_re = re.compile("Parent module '[a-zA-Z0-9]+' not found while handling absolute import")


def run_example_tests():
    global TOTAL, FAILURES, ERRORS
    logging.info('running example tests')
    examples_scripts = list()
    for root, dirs, files in os.walk(os.path.join(os.path.dirname(__file__), "docs", "gallery")):
        for f in files:
            if f.endswith(".py"):
                examples_scripts.append(os.path.join(root, f))

    TOTAL += len(examples_scripts)
    for script in examples_scripts:
        try:
            logging.info("Executing %s" % script)
            ws = list()
            with warnings.catch_warnings(record=True) as tmp:
                _import_from_file(script)
                for w in tmp:  # ignore RunTimeWarnings about importing
                    if isinstance(w.message, RuntimeWarning) and not warning_re.match(str(w.message)):
                        ws.append(w)
            for w in ws:
                warnings.showwarning(w.message, w.category, w.filename, w.lineno, w.line)
        except Exception:
            print(traceback.format_exc())
            FAILURES += 1
            ERRORS += 1


def main():
    # setup and parse arguments
    parser = argparse.ArgumentParser('python test.py [options]')
    parser.set_defaults(verbosity=1, suites=[])
    parser.add_argument('-v', '--verbose', const=2, dest='verbosity', action='store_const', help='run in verbose mode')
    parser.add_argument('-q', '--quiet', const=0, dest='verbosity', action='store_const', help='run disabling output')
    parser.add_argument('-u', '--unit', action='append_const', const=flags['hdmf'], dest='suites',
                        help='run unit tests for hdmf package')
    parser.add_argument('-e', '--example', action='append_const', const=flags['example'], dest='suites',
                        help='run example tests')
    args = parser.parse_args()
    if not args.suites:
        args.suites = list(flags.values())
        args.suites.pop(args.suites.index(flags['example']))  # remove example as a suite run by default

    # set up logger
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter('======================================================================\n'
                                  '%(asctime)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    root.addHandler(ch)

    warnings.simplefilter('always')

    # many tests use NamespaceCatalog.add_namespace, which is deprecated, to set up tests.
    # ignore these warnings for now.
    warnings.filterwarnings("ignore", category=DeprecationWarning, module="hdmf.spec.namespace",
                            message=("NamespaceCatalog.add_namespace has been deprecated. "
                                     "SpecNamespaces should be added with load_namespaces."))

    # Run unit tests for hdmf package
    if flags['hdmf'] in args.suites:
        run_test_suite("tests/unit", "hdmf unit tests", verbose=args.verbosity)

    # Run example tests
    if flags['example'] in args.suites:
        run_example_tests()

    final_message = 'Ran %s tests' % TOTAL
    exitcode = 0
    if ERRORS > 0 or FAILURES > 0:
        exitcode = 1
        _list = list()
        if ERRORS > 0:
            _list.append('errors=%d' % ERRORS)
        if FAILURES > 0:
            _list.append('failures=%d' % FAILURES)
        final_message = '%s - FAILED (%s)' % (final_message, ','.join(_list))
    else:
        final_message = '%s - OK' % final_message

    logging.info(final_message)

    return exitcode


if __name__ == "__main__":
    sys.exit(main())



from .builders import Builder
from .builders import GroupBuilder
from .builders import DatasetBuilder
from .builders import ReferenceBuilder
from .builders import RegionBuilder
from .builders import LinkBuilder

from .objectmapper import ObjectMapper

from .manager import BuildManager
from .manager import TypeMap
